"""
Base class for all Hoppr plug-ins
"""

import functools
import logging
import os
import subprocess
import time
import traceback
from abc import ABC, abstractmethod
from logging.handlers import MemoryHandler
from typing import Any, Callable, List, Optional

from packageurl import PackageURL  # type: ignore

from hoppr import plugin_utils
from hoppr.context import Context
from hoppr.result import Result
from hoppr.utils import obscure_passwords


def _get_component(*args, **kwargs) -> Optional[Any]:
    # TODO: This will only utilize the first Component found in 1) args, 2) kwargs  # pylint: disable=fixme
    # and assumes all plugins operate this way
    comp = None
    for arg in args:
        if type(arg).__name__ == "Component":
            comp = arg
            break

    if comp is None:
        for value in kwargs.values():
            if type(value).__name__ == "Component":
                comp = value
                break

    return comp


def hoppr_process(func: Callable) -> Callable:
    """
    Decorator to handle generic bookkeeping for hoppr plug-ins
    """

    @functools.wraps(func)
    def wrapper(self, *args, **kwargs) -> Result:
        self._start_time = time.time()  # pylint: disable=protected-access
        self.create_logger()

        comp = _get_component(*args, **kwargs)

        arg_string = ""
        if comp is not None:
            arg_string = f"Component purl: {comp.purl}"

        self.get_logger().info(
            f"Starting {self.__class__.__name__}.{func.__name__} {arg_string}"
        )

        result = None
        if comp is not None:
            if comp.purl is None:
                result = Result.fail("No purl supplied for component")

            else:
                purl_type = PackageURL.from_string(comp.purl).type

                if not self.supports_purl_type(purl_type):
                    return Result.skip(
                        f"Class {self.__class__.__name__} does not support purl type {purl_type}"
                    )

                self.get_logger().info(f"Processing purl {comp.purl}")

        if result is None:

            # Only check for missing commands if func has been overridden
            if func.__module__ != HopprPlugin.__module__:
                command_result = plugin_utils.check_for_missing_commands(
                    self.required_commands
                )
                if command_result.is_fail():
                    self.get_logger().error(command_result.message)
                    return command_result

            try:
                result = func(self, *args, **kwargs)
            except Exception as error:  # pylint: disable=broad-except
                self.get_logger().error(
                    "Unexpected exception running %s.%s: %s",
                    self.__class__.__name__,
                    func.__name__,
                    error,
                )
                self.get_logger().error(traceback.format_exc())
                result = Result.fail(f"Unexpected Exception: {error}")

        self.get_logger().info(
            "Completed %s.%s", self.__class__.__name__, func.__name__
        )

        duration = time.time() - self._start_time  # pylint: disable=protected-access
        self.get_logger().info(f"Process duration {duration:3f} seconds")
        self.get_logger().info(f"Result: '{result}'")

        if result.is_skip():
            for handler in self.get_logger().handlers:
                if isinstance(handler, MemoryHandler):
                    handler.setTarget(None)

        self.close_logger()

        return result

    return wrapper


def hoppr_rerunner(method: Callable) -> Callable:
    """
    Runs a method (assumed to return a Result object) a number of times,
    or until the result type is not RETRY
    """

    @functools.wraps(method)
    def wrapper(self, *args, **kwargs) -> Result:

        log = self.get_logger()

        for attempt in range(self.context.max_attempts):
            result = method(self, *args, **kwargs)

            if not isinstance(result, Result):
                msg = (
                    f"Method {method.__name__} returned {type(result).__name__}, "
                    + "in rerunner.  Result object required"
                )
                log.error(msg)
                return Result.fail(msg)

            if not result.is_retry():
                return result

            if attempt < self.context.max_attempts - 1:
                log.warning(
                    f"Method {method.__name__} will be retried in "
                    + f"{self.context.retry_wait_seconds} seconds"
                )
                log.warning(
                    f"  Result message for attempt {attempt + 1}: {result.message}"
                )
                time.sleep(self.context.retry_wait_seconds)

        log.error(
            f"Method {method.__name__} failed after {self.context.max_attempts} attempts",
        )
        log.error(f"  Result message for final attempt: {result.message}")

        return Result.fail(
            f"Failure after {self.context.max_attempts} attempts, final message {result.message}"
        )

    return wrapper


class HopprPlugin(ABC):
    """
    Base class for all Hoppr plug-ins

    Note that this class is not thread safe
    """

    required_commands: List[str] = []
    supported_purl_types: List[str] = []

    def __init__(self, context: Context, config: Any = None) -> None:
        self._logger: logging.Logger
        self._start_time: float = 0.0
        self.context = context
        self.config = config

    @abstractmethod
    def get_version(self) -> str:
        """
        Returns the version of this plug-in
        """

    @hoppr_process
    @hoppr_rerunner
    def pre_stage_process(self) -> Result:
        """
        Process to be run before other processing within a stage for this plug-in
        """

        return Result.skip("pre_stage_process not defined.")

    @hoppr_process
    @hoppr_rerunner
    def process_component(self, comp: Any) -> Result:  # pylint: disable=unused-argument
        """
        Process a single component through this plug-in
        """

        return Result.skip("process_component not defined.")

    @hoppr_process
    @hoppr_rerunner
    def post_stage_process(self):
        """
        Finalize processing for this plug-in
        """
        return Result.skip("post_stage_process not defined.")

    @classmethod
    def supports_purl_type(cls, purl_type: str) -> bool:
        """
        Indicates whether or not this particular plug-in supports components of the specified
        purl type.

        If a plug-in supports all purl types, this method can be overridden to always return True.
        """
        return purl_type in cls.supported_purl_types

    def create_logger(self) -> None:
        """
        Creates a logger with a MemoryHandler, which flushes to stderr when closed
        """

        logger_name: str = f"{self.__class__.__name__}--{os.getpid()}"

        logger = logging.getLogger(logger_name)
        logger.setLevel(logging.INFO)

        formatter = logging.Formatter(
            "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
        )
        stream_handler = logging.StreamHandler()
        stream_handler.setFormatter(formatter)

        log_handler = MemoryHandler(
            10000, flushLevel=logging.CRITICAL, target=stream_handler
        )
        log_handler.setFormatter(formatter)
        logger.addHandler(log_handler)
        self._logger = logger

    def get_logger(self) -> logging.Logger:
        """
        Returns the logger to be used for the current process
        """
        return self._logger

    def close_logger(self) -> None:
        """
        Close (and flush) all handlers for this plug-in's logger
        """
        for handler in self._logger.handlers:
            handler.close()

    def run_command(self, command, password_list=None, cwd=None):
        """
        Run a command and log any errors
        """
        result = subprocess.run(command, check=False, capture_output=True, cwd=cwd)
        if result.returncode != 0:
            obscured_command = obscure_passwords(command, password_list)
            self.get_logger().error(  # pylint: disable=logging-fstring-interpolation
                f"Failed to execute command: {obscured_command}"
            )
            self.get_logger().info(  # pylint: disable=logging-fstring-interpolation
                f"PROCESS STDOUT:\n{result.stdout.decode('utf-8').strip()}"
            )
            self.get_logger().info(  # pylint: disable=logging-fstring-interpolation
                f"PROCESS STDERR:\n{result.stderr.decode('utf-8').strip()}"
            )

        return result



# generated by datamodel-codegen:
#   filename:  bom-1.4.schema.json
#   timestamp: 2022-04-27T20:56:27+00:00

from __future__ import annotations

from enum import Enum


class Schema(Enum):
    AFL_2_0 = "AFL-2.0"
    AAL = "AAL"
    Adobe_2006 = "Adobe-2006"
    AFL_3_0 = "AFL-3.0"
    ADSL = "ADSL"
    field_0BSD = "0BSD"
    Afmparse = "Afmparse"
    AFL_1_2 = "AFL-1.2"
    AGPL_1_0_or_later = "AGPL-1.0-or-later"
    AFL_2_1 = "AFL-2.1"
    AFL_1_1 = "AFL-1.1"
    AGPL_1_0 = "AGPL-1.0"
    Adobe_Glyph = "Adobe-Glyph"
    AMDPLPA = "AMDPLPA"
    Aladdin = "Aladdin"
    ANTLR_PD = "ANTLR-PD"
    AML = "AML"
    Apache_1_0 = "Apache-1.0"
    ANTLR_PD_fallback = "ANTLR-PD-fallback"
    Abstyles = "Abstyles"
    AGPL_1_0_only = "AGPL-1.0-only"
    APAFML = "APAFML"
    APSL_1_0 = "APSL-1.0"
    APSL_1_1 = "APSL-1.1"
    APSL_2_0 = "APSL-2.0"
    AGPL_3_0_only = "AGPL-3.0-only"
    Apache_1_1 = "Apache-1.1"
    Apache_2_0 = "Apache-2.0"
    APL_1_0 = "APL-1.0"
    Bahyph = "Bahyph"
    Artistic_1_0 = "Artistic-1.0"
    AMPAS = "AMPAS"
    Barr = "Barr"
    AGPL_3_0_or_later = "AGPL-3.0-or-later"
    BlueOak_1_0_0 = "BlueOak-1.0.0"
    Beerware = "Beerware"
    Artistic_1_0_cl8 = "Artistic-1.0-cl8"
    blessing = "blessing"
    Borceux = "Borceux"
    BSD_2_Clause_NetBSD = "BSD-2-Clause-NetBSD"
    BSD_1_Clause = "BSD-1-Clause"
    BSD_2_Clause_Patent = "BSD-2-Clause-Patent"
    BitTorrent_1_0 = "BitTorrent-1.0"
    BSD_2_Clause_FreeBSD = "BSD-2-Clause-FreeBSD"
    BSD_3_Clause_Attribution = "BSD-3-Clause-Attribution"
    BSD_2_Clause = "BSD-2-Clause"
    APSL_1_2 = "APSL-1.2"
    BSD_3_Clause_LBNL = "BSD-3-Clause-LBNL"
    Artistic_2_0 = "Artistic-2.0"
    BSD_3_Clause_No_Nuclear_License_2014 = "BSD-3-Clause-No-Nuclear-License-2014"
    BSD_3_Clause_Modification = "BSD-3-Clause-Modification"
    BSD_4_Clause_Shortened = "BSD-4-Clause-Shortened"
    BSD_3_Clause = "BSD-3-Clause"
    BSD_3_Clause_Open_MPI = "BSD-3-Clause-Open-MPI"
    BitTorrent_1_1 = "BitTorrent-1.1"
    BSD_3_Clause_No_Nuclear_Warranty = "BSD-3-Clause-No-Nuclear-Warranty"
    BSD_Source_Code = "BSD-Source-Code"
    BSD_Protection = "BSD-Protection"
    AGPL_3_0 = "AGPL-3.0"
    BUSL_1_1 = "BUSL-1.1"
    Artistic_1_0_Perl = "Artistic-1.0-Perl"
    BSL_1_0 = "BSL-1.0"
    BSD_2_Clause_Views = "BSD-2-Clause-Views"
    CAL_1_0_Combined_Work_Exception = "CAL-1.0-Combined-Work-Exception"
    CATOSL_1_1 = "CATOSL-1.1"
    bzip2_1_0_5 = "bzip2-1.0.5"
    bzip2_1_0_6 = "bzip2-1.0.6"
    CC_BY_2_5 = "CC-BY-2.5"
    CC_BY_3_0_AT = "CC-BY-3.0-AT"
    C_UDA_1_0 = "C-UDA-1.0"
    CC_BY_3_0_US = "CC-BY-3.0-US"
    CC_BY_1_0 = "CC-BY-1.0"
    CC_BY_NC_1_0 = "CC-BY-NC-1.0"
    CC_BY_NC_2_0 = "CC-BY-NC-2.0"
    CC_BY_NC_2_5 = "CC-BY-NC-2.5"
    CC_BY_2_0 = "CC-BY-2.0"
    CC_BY_NC_4_0 = "CC-BY-NC-4.0"
    CC_BY_NC_ND_1_0 = "CC-BY-NC-ND-1.0"
    CC_BY_NC_ND_2_0 = "CC-BY-NC-ND-2.0"
    CC_BY_NC_3_0 = "CC-BY-NC-3.0"
    CC_BY_NC_ND_3_0_IGO = "CC-BY-NC-ND-3.0-IGO"
    CC_BY_NC_ND_3_0 = "CC-BY-NC-ND-3.0"
    BSD_3_Clause_No_Nuclear_License = "BSD-3-Clause-No-Nuclear-License"
    CC_BY_NC_ND_4_0 = "CC-BY-NC-ND-4.0"
    CC_BY_NC_SA_2_0 = "CC-BY-NC-SA-2.0"
    CC_BY_NC_SA_2_5 = "CC-BY-NC-SA-2.5"
    CC_BY_NC_SA_3_0 = "CC-BY-NC-SA-3.0"
    CC_BY_NC_SA_4_0 = "CC-BY-NC-SA-4.0"
    CC_BY_ND_1_0 = "CC-BY-ND-1.0"
    BSD_3_Clause_Clear = "BSD-3-Clause-Clear"
    CC_BY_ND_2_5 = "CC-BY-ND-2.5"
    CC_BY_ND_3_0 = "CC-BY-ND-3.0"
    CC_BY_ND_4_0 = "CC-BY-ND-4.0"
    CC_BY_SA_1_0 = "CC-BY-SA-1.0"
    CC_BY_SA_2_0_UK = "CC-BY-SA-2.0-UK"
    CC_BY_SA_2_0 = "CC-BY-SA-2.0"
    CC_BY_SA_2_1_JP = "CC-BY-SA-2.1-JP"
    CC_BY_ND_2_0 = "CC-BY-ND-2.0"
    CC_BY_SA_3_0_AT = "CC-BY-SA-3.0-AT"
    CC_BY_SA_3_0 = "CC-BY-SA-3.0"
    CC_BY_SA_4_0 = "CC-BY-SA-4.0"
    CC_BY_SA_2_5 = "CC-BY-SA-2.5"
    CC_BY_3_0 = "CC-BY-3.0"
    CDDL_1_0 = "CDDL-1.0"
    CC0_1_0 = "CC0-1.0"
    CC_BY_NC_ND_2_5 = "CC-BY-NC-ND-2.5"
    CC_BY_NC_SA_1_0 = "CC-BY-NC-SA-1.0"
    Caldera = "Caldera"
    CDLA_Permissive_1_0 = "CDLA-Permissive-1.0"
    CC_BY_4_0 = "CC-BY-4.0"
    CC_PDDC = "CC-PDDC"
    BSD_4_Clause_UC = "BSD-4-Clause-UC"
    BSD_4_Clause = "BSD-4-Clause"
    CAL_1_0 = "CAL-1.0"
    CDDL_1_1 = "CDDL-1.1"
    CERN_OHL_1_2 = "CERN-OHL-1.2"
    CERN_OHL_1_1 = "CERN-OHL-1.1"
    CERN_OHL_P_2_0 = "CERN-OHL-P-2.0"
    CERN_OHL_S_2_0 = "CERN-OHL-S-2.0"
    CECILL_1_1 = "CECILL-1.1"
    CECILL_2_0 = "CECILL-2.0"
    CECILL_1_0 = "CECILL-1.0"
    CNRI_Python = "CNRI-Python"
    CNRI_Python_GPL_Compatible = "CNRI-Python-GPL-Compatible"
    copyleft_next_0_3_0 = "copyleft-next-0.3.0"
    CPAL_1_0 = "CPAL-1.0"
    copyleft_next_0_3_1 = "copyleft-next-0.3.1"
    CPL_1_0 = "CPL-1.0"
    ClArtistic = "ClArtistic"
    CECILL_C = "CECILL-C"
    CNRI_Jython = "CNRI-Jython"
    Condor_1_1 = "Condor-1.1"
    CPOL_1_02 = "CPOL-1.02"
    curl = "curl"
    diffmark = "diffmark"
    Crossword = "Crossword"
    Dotseqn = "Dotseqn"
    DOC = "DOC"
    DSDP = "DSDP"
    DRL_1_0 = "DRL-1.0"
    ECL_1_0 = "ECL-1.0"
    ECL_2_0 = "ECL-2.0"
    eCos_2_0 = "eCos-2.0"
    CrystalStacker = "CrystalStacker"
    CERN_OHL_W_2_0 = "CERN-OHL-W-2.0"
    D_FSL_1_0 = "D-FSL-1.0"
    eGenix = "eGenix"
    EPICS = "EPICS"
    Entessa = "Entessa"
    EPL_1_0 = "EPL-1.0"
    EFL_2_0 = "EFL-2.0"
    CUA_OPL_1_0 = "CUA-OPL-1.0"
    etalab_2_0 = "etalab-2.0"
    EUPL_1_0 = "EUPL-1.0"
    ErlPL_1_1 = "ErlPL-1.1"
    EUDatagrid = "EUDatagrid"
    EUPL_1_1 = "EUPL-1.1"
    Cube = "Cube"
    dvipdfm = "dvipdfm"
    FreeBSD_DOC = "FreeBSD-DOC"
    Eurosym = "Eurosym"
    FSFAP = "FSFAP"
    FreeImage = "FreeImage"
    FSFULLR = "FSFULLR"
    FSFUL = "FSFUL"
    GD = "GD"
    GFDL_1_1_invariants_only = "GFDL-1.1-invariants-only"
    EUPL_1_2 = "EUPL-1.2"
    EPL_2_0 = "EPL-2.0"
    GFDL_1_1_no_invariants_or_later = "GFDL-1.1-no-invariants-or-later"
    GFDL_1_1_only = "GFDL-1.1-only"
    GFDL_1_1_or_later = "GFDL-1.1-or-later"
    GFDL_1_1 = "GFDL-1.1"
    FTL = "FTL"
    GFDL_1_2_invariants_or_later = "GFDL-1.2-invariants-or-later"
    GFDL_1_1_invariants_or_later = "GFDL-1.1-invariants-or-later"
    GFDL_1_2_invariants_only = "GFDL-1.2-invariants-only"
    GFDL_1_2_only = "GFDL-1.2-only"
    GFDL_1_2_or_later = "GFDL-1.2-or-later"
    GFDL_1_2 = "GFDL-1.2"
    GFDL_1_2_no_invariants_only = "GFDL-1.2-no-invariants-only"
    GFDL_1_3_invariants_only = "GFDL-1.3-invariants-only"
    GFDL_1_3_no_invariants_only = "GFDL-1.3-no-invariants-only"
    GFDL_1_3_invariants_or_later = "GFDL-1.3-invariants-or-later"
    GFDL_1_2_no_invariants_or_later = "GFDL-1.2-no-invariants-or-later"
    Fair = "Fair"
    Frameworx_1_0 = "Frameworx-1.0"
    Giftware = "Giftware"
    GFDL_1_1_no_invariants_only = "GFDL-1.1-no-invariants-only"
    GL2PS = "GL2PS"
    Glulxe = "Glulxe"
    Glide = "Glide"
    gnuplot = "gnuplot"
    GLWTPL = "GLWTPL"
    GPL_1_0_only = "GPL-1.0-only"
    GPL_1_0_or_later = "GPL-1.0-or-later"
    GPL_1_0 = "GPL-1.0"
    GFDL_1_3_only = "GFDL-1.3-only"
    GPL_2_0_only = "GPL-2.0-only"
    GPL_2_0_or_later = "GPL-2.0-or-later"
    GPL_2_0_with_autoconf_exception = "GPL-2.0-with-autoconf-exception"
    GPL_2_0_ = "GPL-2.0+"
    GFDL_1_3 = "GFDL-1.3"
    GPL_1_0_ = "GPL-1.0+"
    CDLA_Sharing_1_0 = "CDLA-Sharing-1.0"
    GPL_2_0_with_classpath_exception = "GPL-2.0-with-classpath-exception"
    GPL_2_0_with_GCC_exception = "GPL-2.0-with-GCC-exception"
    GPL_2_0_with_bison_exception = "GPL-2.0-with-bison-exception"
    GPL_2_0_with_font_exception = "GPL-2.0-with-font-exception"
    GPL_2_0 = "GPL-2.0"
    CECILL_2_1 = "CECILL-2.1"
    GPL_3_0 = "GPL-3.0"
    GPL_3_0_ = "GPL-3.0+"
    GPL_3_0_with_autoconf_exception = "GPL-3.0-with-autoconf-exception"
    GPL_3_0_only = "GPL-3.0-only"
    Hippocratic_2_1 = "Hippocratic-2.1"
    HPND = "HPND"
    HTMLTIDY = "HTMLTIDY"
    GPL_3_0_with_GCC_exception = "GPL-3.0-with-GCC-exception"
    HaskellReport = "HaskellReport"
    GPL_3_0_or_later = "GPL-3.0-or-later"
    ICU = "ICU"
    ImageMagick = "ImageMagick"
    iMatix = "iMatix"
    IBM_pibs = "IBM-pibs"
    Intel_ACPI = "Intel-ACPI"
    Intel = "Intel"
    Info_ZIP = "Info-ZIP"
    IPA = "IPA"
    IJG = "IJG"
    ISC = "ISC"
    JasPer_2_0 = "JasPer-2.0"
    JPNIC = "JPNIC"
    JSON = "JSON"
    LAL_1_2 = "LAL-1.2"
    LAL_1_3 = "LAL-1.3"
    Latex2e = "Latex2e"
    Leptonica = "Leptonica"
    HPND_sell_variant = "HPND-sell-variant"
    LGPL_2_0_only = "LGPL-2.0-only"
    LGPL_2_0_or_later = "LGPL-2.0-or-later"
    Imlib2 = "Imlib2"
    IPL_1_0 = "IPL-1.0"
    LGPL_2_1_only = "LGPL-2.1-only"
    LGPL_2_1_or_later = "LGPL-2.1-or-later"
    LGPL_2_0_ = "LGPL-2.0+"
    LGPL_2_0 = "LGPL-2.0"
    CECILL_B = "CECILL-B"
    LGPL_3_0_or_later = "LGPL-3.0-or-later"
    LGPL_3_0_only = "LGPL-3.0-only"
    LGPLLR = "LGPLLR"
    libpng_2_0 = "libpng-2.0"
    Libpng = "Libpng"
    libselinux_1_0 = "libselinux-1.0"
    LGPL_3_0_ = "LGPL-3.0+"
    EFL_1_0 = "EFL-1.0"
    libtiff = "libtiff"
    GFDL_1_3_no_invariants_or_later = "GFDL-1.3-no-invariants-or-later"
    LiLiQ_Rplus_1_1 = "LiLiQ-Rplus-1.1"
    LiLiQ_R_1_1 = "LiLiQ-R-1.1"
    LPL_1_0 = "LPL-1.0"
    LiLiQ_P_1_1 = "LiLiQ-P-1.1"
    Linux_OpenIB = "Linux-OpenIB"
    LPPL_1_0 = "LPPL-1.0"
    LPPL_1_2 = "LPPL-1.2"
    LPPL_1_3a = "LPPL-1.3a"
    LPL_1_02 = "LPL-1.02"
    LPPL_1_3c = "LPPL-1.3c"
    MakeIndex = "MakeIndex"
    LGPL_2_1_ = "LGPL-2.1+"
    LPPL_1_1 = "LPPL-1.1"
    MIT_CMU = "MIT-CMU"
    MirOS = "MirOS"
    MIT_advertising = "MIT-advertising"
    MIT_Modern_Variant = "MIT-Modern-Variant"
    MIT = "MIT"
    MIT_enna = "MIT-enna"
    MIT_open_group = "MIT-open-group"
    MIT_feh = "MIT-feh"
    MITNFA = "MITNFA"
    MPL_1_0 = "MPL-1.0"
    mpich2 = "mpich2"
    MPL_2_0 = "MPL-2.0"
    MPL_2_0_no_copyleft_exception = "MPL-2.0-no-copyleft-exception"
    MS_RL = "MS-RL"
    MTLL = "MTLL"
    MPL_1_1 = "MPL-1.1"
    MulanPSL_2_0 = "MulanPSL-2.0"
    Motosoto = "Motosoto"
    Mup = "Mup"
    MulanPSL_1_0 = "MulanPSL-1.0"
    NAIST_2003 = "NAIST-2003"
    Naumen = "Naumen"
    Multics = "Multics"
    NBPL_1_0 = "NBPL-1.0"
    NCSA = "NCSA"
    Net_SNMP = "Net-SNMP"
    NetCDF = "NetCDF"
    NASA_1_3 = "NASA-1.3"
    NGPL = "NGPL"
    NIST_PD_fallback = "NIST-PD-fallback"
    NIST_PD = "NIST-PD"
    Newsletr = "Newsletr"
    NLPL = "NLPL"
    Nokia = "Nokia"
    NOSL = "NOSL"
    Noweb = "Noweb"
    NLOD_1_0 = "NLOD-1.0"
    NPL_1_0 = "NPL-1.0"
    NCGL_UK_2_0 = "NCGL-UK-2.0"
    NRL = "NRL"
    NTP_0 = "NTP-0"
    NTP = "NTP"
    GFDL_1_3_or_later = "GFDL-1.3-or-later"
    Nunit = "Nunit"
    O_UDA_1_0 = "O-UDA-1.0"
    NPL_1_1 = "NPL-1.1"
    OCCT_PL = "OCCT-PL"
    ODC_By_1_0 = "ODC-By-1.0"
    OFL_1_0_no_RFN = "OFL-1.0-no-RFN"
    OCLC_2_0 = "OCLC-2.0"
    OFL_1_0_RFN = "OFL-1.0-RFN"
    OFL_1_0 = "OFL-1.0"
    OFL_1_1_no_RFN = "OFL-1.1-no-RFN"
    OFL_1_1_RFN = "OFL-1.1-RFN"
    OFL_1_1 = "OFL-1.1"
    OGDL_Taiwan_1_0 = "OGDL-Taiwan-1.0"
    OGC_1_0 = "OGC-1.0"
    OGL_UK_1_0 = "OGL-UK-1.0"
    OGL_UK_2_0 = "OGL-UK-2.0"
    OGL_UK_3_0 = "OGL-UK-3.0"
    OGTSL = "OGTSL"
    OLDAP_1_1 = "OLDAP-1.1"
    OLDAP_1_2 = "OLDAP-1.2"
    OLDAP_1_3 = "OLDAP-1.3"
    OGL_Canada_2_0 = "OGL-Canada-2.0"
    OLDAP_2_0_1 = "OLDAP-2.0.1"
    OLDAP_2_0 = "OLDAP-2.0"
    OLDAP_2_1 = "OLDAP-2.1"
    OLDAP_2_2_1 = "OLDAP-2.2.1"
    OLDAP_2_2_2 = "OLDAP-2.2.2"
    OLDAP_2_2 = "OLDAP-2.2"
    ODbL_1_0 = "ODbL-1.0"
    OLDAP_2_4 = "OLDAP-2.4"
    OLDAP_1_4 = "OLDAP-1.4"
    OLDAP_2_3 = "OLDAP-2.3"
    OLDAP_2_7 = "OLDAP-2.7"
    OLDAP_2_8 = "OLDAP-2.8"
    OML = "OML"
    OpenSSL = "OpenSSL"
    OLDAP_2_6 = "OLDAP-2.6"
    OPL_1_0 = "OPL-1.0"
    OSL_1_0 = "OSL-1.0"
    OSL_1_1 = "OSL-1.1"
    OSL_2_0 = "OSL-2.0"
    OSET_PL_2_1 = "OSET-PL-2.1"
    OSL_2_1 = "OSL-2.1"
    Parity_6_0_0 = "Parity-6.0.0"
    Parity_7_0_0 = "Parity-7.0.0"
    PDDL_1_0 = "PDDL-1.0"
    PHP_3_0 = "PHP-3.0"
    OSL_3_0 = "OSL-3.0"
    Plexus = "Plexus"
    MS_PL = "MS-PL"
    PolyForm_Small_Business_1_0_0 = "PolyForm-Small-Business-1.0.0"
    PolyForm_Noncommercial_1_0_0 = "PolyForm-Noncommercial-1.0.0"
    PSF_2_0 = "PSF-2.0"
    psfrag = "psfrag"
    PostgreSQL = "PostgreSQL"
    psutils = "psutils"
    Qhull = "Qhull"
    QPL_1_0 = "QPL-1.0"
    Rdisc = "Rdisc"
    Python_2_0 = "Python-2.0"
    RPL_1_1 = "RPL-1.1"
    RPL_1_5 = "RPL-1.5"
    RHeCos_1_1 = "RHeCos-1.1"
    RSA_MD = "RSA-MD"
    RSCPL = "RSCPL"
    Ruby = "Ruby"
    SAX_PD = "SAX-PD"
    Saxpath = "Saxpath"
    SCEA = "SCEA"
    Sendmail_8_23 = "Sendmail-8.23"
    Sendmail = "Sendmail"
    SGI_B_1_0 = "SGI-B-1.0"
    SGI_B_1_1 = "SGI-B-1.1"
    SGI_B_2_0 = "SGI-B-2.0"
    SHL_0_5 = "SHL-0.5"
    SHL_0_51 = "SHL-0.51"
    SimPL_2_0 = "SimPL-2.0"
    SISSL_1_2 = "SISSL-1.2"
    SISSL = "SISSL"
    Sleepycat = "Sleepycat"
    SMLNJ = "SMLNJ"
    SMPPL = "SMPPL"
    SNIA = "SNIA"
    Spencer_86 = "Spencer-86"
    Spencer_94 = "Spencer-94"
    Spencer_99 = "Spencer-99"
    SPL_1_0 = "SPL-1.0"
    SSH_OpenSSH = "SSH-OpenSSH"
    PHP_3_01 = "PHP-3.01"
    SSH_short = "SSH-short"
    MIT_0 = "MIT-0"
    RPSL_1_0 = "RPSL-1.0"
    SWL = "SWL"
    SugarCRM_1_1_3 = "SugarCRM-1.1.3"
    TCL = "TCL"
    TCP_wrappers = "TCP-wrappers"
    SSPL_1_0 = "SSPL-1.0"
    TMate = "TMate"
    TOSL = "TOSL"
    TORQUE_1_1 = "TORQUE-1.1"
    TAPR_OHL_1_0 = "TAPR-OHL-1.0"
    UCL_1_0 = "UCL-1.0"
    Unicode_DFS_2015 = "Unicode-DFS-2015"
    Unicode_DFS_2016 = "Unicode-DFS-2016"
    Unicode_TOU = "Unicode-TOU"
    TU_Berlin_1_0 = "TU-Berlin-1.0"
    UPL_1_0 = "UPL-1.0"
    Unlicense = "Unlicense"
    VOSTROM = "VOSTROM"
    Vim = "Vim"
    VSL_1_0 = "VSL-1.0"
    W3C_20150513 = "W3C-20150513"
    W3C = "W3C"
    W3C_19980720 = "W3C-19980720"
    Wsuipa = "Wsuipa"
    Watcom_1_0 = "Watcom-1.0"
    WTFPL = "WTFPL"
    X11 = "X11"
    Xerox = "Xerox"
    XFree86_1_1 = "XFree86-1.1"
    xinetd = "xinetd"
    Xnet = "Xnet"
    xpp = "xpp"
    XSkat = "XSkat"
    YPL_1_0 = "YPL-1.0"
    YPL_1_1 = "YPL-1.1"
    Zed = "Zed"
    Zend_2_0 = "Zend-2.0"
    TU_Berlin_2_0 = "TU-Berlin-2.0"
    Zimbra_1_4 = "Zimbra-1.4"
    zlib_acknowledgement = "zlib-acknowledgement"
    Zlib = "Zlib"
    ZPL_1_1 = "ZPL-1.1"
    ZPL_2_0 = "ZPL-2.0"
    ZPL_2_1 = "ZPL-2.1"
    wxWindows = "wxWindows"
    Zimbra_1_3 = "Zimbra-1.3"
    gSOAP_1_3b = "gSOAP-1.3b"
    Interbase_1_0 = "Interbase-1.0"
    LGPL_2_1 = "LGPL-2.1"
    LGPL_3_0 = "LGPL-3.0"
    NPOSL_3_0 = "NPOSL-3.0"
    OLDAP_2_5 = "OLDAP-2.5"
    StandardML_NJ = "StandardML-NJ"
    field_389_exception = "389-exception"
    Autoconf_exception_2_0 = "Autoconf-exception-2.0"
    Autoconf_exception_3_0 = "Autoconf-exception-3.0"
    Bison_exception_2_2 = "Bison-exception-2.2"
    Bootloader_exception = "Bootloader-exception"
    Classpath_exception_2_0 = "Classpath-exception-2.0"
    CLISP_exception_2_0 = "CLISP-exception-2.0"
    DigiRule_FOSS_exception = "DigiRule-FOSS-exception"
    eCos_exception_2_0 = "eCos-exception-2.0"
    Fawkes_Runtime_exception = "Fawkes-Runtime-exception"
    FLTK_exception = "FLTK-exception"
    Font_exception_2_0 = "Font-exception-2.0"
    freertos_exception_2_0 = "freertos-exception-2.0"
    GCC_exception_2_0 = "GCC-exception-2.0"
    GCC_exception_3_1 = "GCC-exception-3.1"
    gnu_javamail_exception = "gnu-javamail-exception"
    GPL_3_0_linking_exception = "GPL-3.0-linking-exception"
    GPL_3_0_linking_source_exception = "GPL-3.0-linking-source-exception"
    GPL_CC_1_0 = "GPL-CC-1.0"
    i2p_gpl_java_exception = "i2p-gpl-java-exception"
    LGPL_3_0_linking_exception = "LGPL-3.0-linking-exception"
    Libtool_exception = "Libtool-exception"
    Linux_syscall_note = "Linux-syscall-note"
    LLVM_exception = "LLVM-exception"
    LZMA_exception = "LZMA-exception"
    mif_exception = "mif-exception"
    Nokia_Qt_exception_1_1 = "Nokia-Qt-exception-1.1"
    OCaml_LGPL_linking_exception = "OCaml-LGPL-linking-exception"
    OCCT_exception_1_0 = "OCCT-exception-1.0"
    OpenJDK_assembly_exception_1_0 = "OpenJDK-assembly-exception-1.0"
    openvpn_openssl_exception = "openvpn-openssl-exception"
    PS_or_PDF_font_exception_20170817 = "PS-or-PDF-font-exception-20170817"
    Qt_GPL_exception_1_0 = "Qt-GPL-exception-1.0"
    Qt_LGPL_exception_1_1 = "Qt-LGPL-exception-1.1"
    Qwt_exception_1_0 = "Qwt-exception-1.0"
    SHL_2_0 = "SHL-2.0"
    SHL_2_1 = "SHL-2.1"
    Swift_exception = "Swift-exception"
    u_boot_exception_2_0 = "u-boot-exception-2.0"
    Universal_FOSS_exception_1_0 = "Universal-FOSS-exception-1.0"
    WxWindows_exception_3_1 = "WxWindows-exception-3.1"



"""
Define an object for a single set of credentials
"""

from dataclasses import dataclass


@dataclass
class CredObject:
    """
    Define credential details for a single set of credentials
    """

    username: str
    password: str



# generated by datamodel-codegen:
#   filename:  bom-1.3.schema.json
#   timestamp: 2022-04-27T20:56:53+00:00

from __future__ import annotations

from datetime import datetime
from enum import Enum
from typing import Any, List, Optional, Union

from pydantic import BaseModel, EmailStr, Field, constr

from . import spdx


class BomFormat(Enum):
    CycloneDX = "CycloneDX"


class OrganizationalContact(BaseModel):
    name: Optional[str] = Field(
        None,
        description="The name of a contact",
        examples=["Contact name"],
        title="Name",
    )
    email: Optional[str] = Field(
        None,
        description="The email address of the contact.",
        examples=["firstname.lastname@example.com"],
        title="Email Address",
    )
    phone: Optional[str] = Field(
        None,
        description="The phone number of the contact.",
        examples=["800-555-1212"],
        title="Phone",
    )


class Type(Enum):
    application = "application"
    framework = "framework"
    library = "library"
    container = "container"
    operating_system = "operating-system"
    device = "device"
    firmware = "firmware"
    file = "file"


class Scope(Enum):
    required = "required"
    optional = "optional"
    excluded = "excluded"


class Encoding(Enum):
    base64 = "base64"


class Attachment(BaseModel):
    contentType: Optional[str] = Field(
        "text/plain",
        description="Specifies the content type of the text. Defaults to text/plain if not specified.",
        title="Content-Type",
    )
    encoding: Optional[Encoding] = Field(
        None,
        description="Specifies the optional encoding the text is represented in.",
        title="Encoding",
    )
    content: str = Field(
        ..., description="The attachment data", title="Attachment Text"
    )


class HashAlg(Enum):
    MD5 = "MD5"
    SHA_1 = "SHA-1"
    SHA_256 = "SHA-256"
    SHA_384 = "SHA-384"
    SHA_512 = "SHA-512"
    SHA3_256 = "SHA3-256"
    SHA3_384 = "SHA3-384"
    SHA3_512 = "SHA3-512"
    BLAKE2b_256 = "BLAKE2b-256"
    BLAKE2b_384 = "BLAKE2b-384"
    BLAKE2b_512 = "BLAKE2b-512"
    BLAKE3 = "BLAKE3"


class HashContent(BaseModel):
    __root__: constr(
        regex=r"^([a-fA-F0-9]{32}|[a-fA-F0-9]{40}|[a-fA-F0-9]{64}|[a-fA-F0-9]{96}|[a-fA-F0-9]{128})$"
    ) = Field(
        ..., examples=["3942447fac867ae5cdb3229b658f4d48"], title="Hash Content (value)"
    )


class Type1(Enum):
    unofficial = "unofficial"
    monkey = "monkey"
    backport = "backport"
    cherry_pick = "cherry-pick"


class Diff(BaseModel):
    text: Optional[Attachment] = Field(
        None, description="Specifies the optional text of the diff", title="Diff text"
    )
    url: Optional[str] = Field(
        None, description="Specifies the URL to the diff", title="URL"
    )


class Type2(Enum):
    defect = "defect"
    enhancement = "enhancement"
    security = "security"


class Source(BaseModel):
    name: Optional[str] = Field(
        None,
        description="The name of the source. For example 'National Vulnerability Database', 'NVD', and 'Apache'",
        title="Name",
    )
    url: Optional[str] = Field(
        None,
        description="The url of the issue documentation as provided by the source",
        title="URL",
    )


class Issue(BaseModel):
    type: Type2 = Field(..., description="Specifies the type of issue", title="Type")
    id: Optional[str] = Field(
        None,
        description="The identifier of the issue assigned by the source of the issue",
        title="ID",
    )
    name: Optional[str] = Field(None, description="The name of the issue", title="Name")
    description: Optional[str] = Field(
        None, description="A description of the issue", title="Description"
    )
    source: Optional[Source] = Field(
        None,
        description="The source of the issue where it is documented",
        title="Source",
    )
    references: Optional[List[str]] = Field(
        None,
        description="A collection of URL's for reference. Multiple URLs are allowed.",
        examples=["https://example.com"],
        title="References",
    )


class IdentifiableAction(BaseModel):
    timestamp: Optional[datetime] = Field(
        None,
        description="The timestamp in which the action occurred",
        title="Timestamp",
    )
    name: Optional[str] = Field(
        None,
        description="The name of the individual who performed the action",
        title="Name",
    )
    email: Optional[EmailStr] = Field(
        None,
        description="The email address of the individual who performed the action",
        title="E-mail",
    )


class Type3(Enum):
    vcs = "vcs"
    issue_tracker = "issue-tracker"
    website = "website"
    advisories = "advisories"
    bom = "bom"
    mailing_list = "mailing-list"
    social = "social"
    chat = "chat"
    documentation = "documentation"
    support = "support"
    distribution = "distribution"
    license = "license"
    build_meta = "build-meta"
    build_system = "build-system"
    other = "other"


class Dependency(BaseModel):
    ref: str = Field(
        ...,
        description="References a component by the components bom-ref attribute",
        title="Reference",
    )
    dependsOn: Optional[List[str]] = Field(
        None,
        description="The bom-ref identifiers of the components that are dependencies of this dependency object.",
        title="Depends On",
        # REMOVED IDH unique_items=True,
    )


class DataFlow(Enum):
    inbound = "inbound"
    outbound = "outbound"
    bi_directional = "bi-directional"
    unknown = "unknown"


class Copyright(BaseModel):
    text: str = Field(..., title="Copyright Text")


class AggregateType(Enum):
    complete = "complete"
    incomplete = "incomplete"
    incomplete_first_party_only = "incomplete_first_party_only"
    incomplete_third_party_only = "incomplete_third_party_only"
    unknown = "unknown"
    not_specified = "not_specified"


class Property(BaseModel):
    name: Optional[str] = Field(
        None,
        description="The name of the property. Duplicate names are allowed, each potentially having a different value.",
        title="Name",
    )
    value: Optional[str] = Field(
        None, description="The value of the property.", title="Value"
    )


class OrganizationalEntity(BaseModel):
    name: Optional[str] = Field(
        None,
        description="The name of the organization",
        examples=["Example Inc."],
        title="Name",
    )
    url: Optional[List[str]] = Field(
        None,
        description="The URL of the organization. Multiple URLs are allowed.",
        examples=["https://example.com"],
        title="URL",
    )
    contact: Optional[List[OrganizationalContact]] = Field(
        None,
        description="A contact at the organization. Multiple contacts are allowed.",
        title="Contact",
    )


class Swid(BaseModel):
    tagId: str = Field(
        ..., description="Maps to the tagId of a SoftwareIdentity.", title="Tag ID"
    )
    name: str = Field(
        ..., description="Maps to the name of a SoftwareIdentity.", title="Name"
    )
    version: Optional[str] = Field(
        "0.0", description="Maps to the version of a SoftwareIdentity.", title="Version"
    )
    tagVersion: Optional[int] = Field(
        0,
        description="Maps to the tagVersion of a SoftwareIdentity.",
        title="Tag Version",
    )
    patch: Optional[bool] = Field(
        False, description="Maps to the patch of a SoftwareIdentity.", title="Patch"
    )
    text: Optional[Attachment] = Field(
        None,
        description="Specifies the metadata and content of the SWID tag.",
        title="Attachment text",
    )
    url: Optional[str] = Field(
        None, description="The URL to the SWID file.", title="URL"
    )


class Hash(BaseModel):
    alg: HashAlg
    content: HashContent


class License(BaseModel):
    id: Optional[spdx.Schema] = Field(
        None,
        description="A valid SPDX license ID",
        examples=["Apache-2.0"],
        title="License ID (SPDX)",
    )
    name: Optional[str] = Field(
        None,
        description="If SPDX does not define the license used, this field may be used to provide the license name",
        examples=["Acme Software License"],
        title="License Name",
    )
    text: Optional[Attachment] = Field(
        None,
        description="An optional way to include the textual content of a license.",
        title="License text",
    )
    url: Optional[str] = Field(
        None,
        description="The URL to the license file. If specified, a 'license' externalReference should also be specified for completeness",
        examples=["https://www.apache.org/licenses/LICENSE-2.0.txt"],
        title="License URL",
    )


class License1(BaseModel):
    __root__: Union[License, Any, Any] = Field(..., title="License Object")


class LicenseChoice1(BaseModel):
    license: Optional[License1] = None
    expression: Optional[str] = Field(
        None,
        examples=[
            "Apache-2.0 AND (MIT OR GPL-2.0-only)",
            "GPL-3.0-only WITH Classpath-exception-2.0",
        ],
        title="SPDX License Expression",
    )


class LicenseChoice(BaseModel):
    __root__: Union[LicenseChoice1, Any, Any] = Field(..., title="License(s)")


class Commit(BaseModel):
    uid: Optional[str] = Field(
        None,
        description="A unique identifier of the commit. This may be version control specific. For example, Subversion uses revision numbers whereas git uses commit hashes.",
        title="UID",
    )
    url: Optional[str] = Field(
        None,
        description="The URL to the commit. This URL will typically point to a commit in a version control system.",
        title="URL",
    )
    author: Optional[IdentifiableAction] = Field(
        None,
        description="The author who created the changes in the commit",
        title="Author",
    )
    committer: Optional[IdentifiableAction] = Field(
        None,
        description="The person who committed or pushed the commit",
        title="Committer",
    )
    message: Optional[str] = Field(
        None,
        description="The text description of the contents of the commit",
        title="Message",
    )


class Patch(BaseModel):
    type: Type1 = Field(
        ...,
        description="Specifies the purpose for the patch including the resolution of defects, security issues, or new behavior or functionality",
        title="Type",
    )
    diff: Optional[Diff] = Field(
        None,
        description="The patch file (or diff) that show changes. Refer to https://en.wikipedia.org/wiki/Diff",
        title="Diff",
    )
    resolves: Optional[List[Issue]] = Field(
        None, description="A collection of issues the patch resolves", title="Resolves"
    )


class ExternalReference(BaseModel):
    url: str = Field(..., description="The URL to the external reference", title="URL")
    comment: Optional[str] = Field(
        None,
        description="An optional comment describing the external reference",
        title="Comment",
    )
    type: Type3 = Field(
        ...,
        description='Specifies the type of external reference. There are built-in types to describe common references. If a type does not exist for the reference being referred to, use the "other" type.',
        title="Type",
    )
    hashes: Optional[List[Hash]] = Field(
        None,
        description="The hashes of the external reference (if applicable).",
        title="Hashes",
    )


class DataClassification(BaseModel):
    flow: DataFlow
    classification: str


class ComponentEvidence(BaseModel):
    licenses: Optional[List[LicenseChoice]] = Field(None, title="Component License(s)")
    copyright: Optional[List[Copyright]] = Field(None, title="Copyright")


class Compositions(BaseModel):
    aggregate: AggregateType = Field(
        ...,
        description="Specifies an aggregate type that describe how complete a relationship is.",
        title="Aggregate",
    )
    assemblies: Optional[List[str]] = Field(
        None,
        description="The bom-ref identifiers of the components or services being described. Assemblies refer to nested relationships whereby a constituent part may include other constituent parts. References do not cascade to child parts. References are explicit for the specified constituent part only.",
        title="BOM references",
        # REMOVED IDH unique_items=True,
    )
    dependencies: Optional[List[str]] = Field(
        None,
        description="The bom-ref identifiers of the components or services being described. Dependencies refer to a relationship whereby an independent constituent part requires another independent constituent part. References do not cascade to transitive dependencies. References are explicit for the specified dependency only.",
        title="BOM references",
        # REMOVED IDH unique_items=True,
    )


class Tool(BaseModel):
    vendor: Optional[str] = Field(
        None,
        description="The date and time (timestamp) when the document was created.",
        title="Tool Vendor",
    )
    name: Optional[str] = Field(
        None,
        description="The date and time (timestamp) when the document was created.",
        title="Tool Name",
    )
    version: Optional[str] = Field(
        None,
        description="The date and time (timestamp) when the document was created.",
        title="Tool Version",
    )
    hashes: Optional[List[Hash]] = Field(
        None, description="The hashes of the tool (if applicable).", title="Hashes"
    )


class Service(BaseModel):
    bom_ref: Optional[str] = Field(
        None,
        alias="bom-ref",
        description="An optional identifier which can be used to reference the service elsewhere in the BOM. Every bom-ref should be unique.",
        title="BOM Reference",
    )
    provider: Optional[OrganizationalEntity] = Field(
        None,
        description="The organization that provides the service.",
        title="Provider",
    )
    group: Optional[str] = Field(
        None,
        description="The grouping name, namespace, or identifier. This will often be a shortened, single name of the company or project that produced the service or domain name. Whitespace and special characters should be avoided.",
        examples=["com.acme"],
        title="Service Group",
    )
    name: str = Field(
        ...,
        description="The name of the service. This will often be a shortened, single name of the service.",
        examples=["ticker-service"],
        title="Service Name",
    )
    version: Optional[str] = Field(
        None,
        description="The service version.",
        examples=["1.0.0"],
        title="Service Version",
    )
    description: Optional[str] = Field(
        None,
        description="Specifies a description for the service",
        title="Service Description",
    )
    endpoints: Optional[List[str]] = Field(
        None,
        description="The endpoint URIs of the service. Multiple endpoints are allowed.",
        examples=["https://example.com/api/v1/ticker"],
        title="Endpoints",
    )
    authenticated: Optional[bool] = Field(
        None,
        description="A boolean value indicating if the service requires authentication. A value of true indicates the service requires authentication prior to use. A value of false indicates the service does not require authentication.",
        title="Authentication Required",
    )
    x_trust_boundary: Optional[bool] = Field(
        None,
        alias="x-trust-boundary",
        description="A boolean value indicating if use of the service crosses a trust zone or boundary. A value of true indicates that by using the service, a trust boundary is crossed. A value of false indicates that by using the service, a trust boundary is not crossed.",
        title="Crosses Trust Boundary",
    )
    data: Optional[List[DataClassification]] = Field(
        None,
        description="Specifies the data classification.",
        title="Data Classification",
    )
    licenses: Optional[List[LicenseChoice]] = Field(None, title="Component License(s)")
    externalReferences: Optional[List[ExternalReference]] = Field(
        None, title="External References"
    )
    services: Optional[List[Service]] = Field(None, title="Services")
    properties: Optional[List[Property]] = Field(
        None,
        description="Provides the ability to document properties in a name-value store. This provides flexibility to include data not officially supported in the standard without having to use additional namespaces or create extensions. Unlike key-value stores, properties support duplicate names, each potentially having different values.",
        title="Properties",
    )


class CyclonedxSoftwareBillOfMaterialSpecification(BaseModel):
    bomFormat: BomFormat = Field(
        ...,
        description="Specifies the format of the BOM. This helps to identify the file as CycloneDX since BOMs do not have a filename convention nor does JSON schema support namespaces.",
        title="BOM Format",
    )
    specVersion: str = Field(
        ...,
        description="The version of the CycloneDX specification a BOM is written to (starting at version 1.2)",
        examples=["1.3"],
        title="CycloneDX Specification Version",
    )
    serialNumber: Optional[
        constr(
            regex=r"^urn:uuid:[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$"
        )
    ] = Field(
        None,
        description="Every BOM generated should have a unique serial number, even if the contents of the BOM being generated have not changed over time. The process or tool responsible for creating the BOM should create random UUID's for every BOM generated.",
        examples=["urn:uuid:3e671687-395b-41f5-a30f-a58921a69b79"],
        title="BOM Serial Number",
    )
    version: int = Field(
        ...,
        description="The version allows component publishers/authors to make changes to existing BOMs to update various aspects of the document such as description or licenses. When a system is presented with multiple BOMs for the same component, the system should use the most recent version of the BOM. The default version is '1' and should be incremented for each version of the BOM that is published. Each version of a component should have a unique BOM and if no changes are made to the BOMs, then each BOM will have a version of '1'.",
        examples=[1],
        title="BOM Version",
    )
    metadata: Optional[Metadata] = Field(
        None,
        description="Provides additional information about a BOM.",
        title="BOM Metadata",
    )
    components: Optional[List[Component]] = Field(
        None,
        title="Components",
        # REMOVED IDH unique_items=True,
    )
    services: Optional[List[Service]] = Field(
        None,
        title="Services",
        # REMOVED IDH unique_items=True
    )
    externalReferences: Optional[List[ExternalReference]] = Field(
        None,
        description="External references provide a way to document systems, sites, and information that may be relevant but which are not included with the BOM.",
        title="External References",
    )
    dependencies: Optional[List[Dependency]] = Field(
        None,
        description="Provides the ability to document dependency relationships.",
        title="Dependencies",
        # REMOVED IDH unique_items=True,
    )
    compositions: Optional[List[Compositions]] = Field(
        None,
        description="Compositions describe constituent parts (including components, services, and dependency relationships) and their completeness.",
        title="Compositions",
        # REMOVED IDH unique_items=True,
    )


class Metadata(BaseModel):
    timestamp: Optional[datetime] = Field(
        None,
        description="The date and time (timestamp) when the document was created.",
        title="Timestamp",
    )
    tools: Optional[List[Tool]] = Field(
        None,
        description="The tool(s) used in the creation of the BOM.",
        title="Creation Tools",
    )
    authors: Optional[List[OrganizationalContact]] = Field(
        None,
        description="The person(s) who created the BOM. Authors are common in BOMs created through manual processes. BOMs created through automated means may not have authors.",
        title="Authors",
    )
    component: Optional[Component] = Field(
        None, description="The component that the BOM describes.", title="Component"
    )
    manufacture: Optional[OrganizationalEntity] = Field(
        None,
        description="The organization that manufactured the component that the BOM describes.",
        title="Manufacture",
    )
    supplier: Optional[OrganizationalEntity] = Field(
        None,
        description=" The organization that supplied the component that the BOM describes. The supplier may often be the manufacturer, but may also be a distributor or repackager.",
        title="Supplier",
    )
    licenses: Optional[List[LicenseChoice]] = Field(None, title="BOM License(s)")
    properties: Optional[List[Property]] = Field(
        None,
        description="Provides the ability to document properties in a name-value store. This provides flexibility to include data not officially supported in the standard without having to use additional namespaces or create extensions. Unlike key-value stores, properties support duplicate names, each potentially having different values.",
        title="Properties",
    )


class Pedigree(BaseModel):
    ancestors: Optional[List[Component]] = Field(
        None,
        description="Describes zero or more components in which a component is derived from. This is commonly used to describe forks from existing projects where the forked version contains a ancestor node containing the original component it was forked from. For example, Component A is the original component. Component B is the component being used and documented in the BOM. However, Component B contains a pedigree node with a single ancestor documenting Component A - the original component from which Component B is derived from.",
        title="Ancestors",
    )
    descendants: Optional[List[Component]] = Field(
        None,
        description="Descendants are the exact opposite of ancestors. This provides a way to document all forks (and their forks) of an original or root component.",
        title="Descendants",
    )
    variants: Optional[List[Component]] = Field(
        None,
        description="Variants describe relations where the relationship between the components are not known. For example, if Component A contains nearly identical code to Component B. They are both related, but it is unclear if one is derived from the other, or if they share a common ancestor.",
        title="Variants",
    )
    commits: Optional[List[Commit]] = Field(
        None,
        description="A list of zero or more commits which provide a trail describing how the component deviates from an ancestor, descendant, or variant.",
        title="Commits",
    )
    patches: Optional[List[Patch]] = Field(
        None,
        description=">A list of zero or more patches describing how the component deviates from an ancestor, descendant, or variant. Patches may be complimentary to commits or may be used in place of commits.",
        title="Patches",
    )
    notes: Optional[str] = Field(
        None,
        description="Notes, observations, and other non-structured commentary describing the components pedigree.",
        title="Notes",
    )


class Component(BaseModel):
    type: Type = Field(
        ...,
        description="Specifies the type of component. For software components, classify as application if no more specific appropriate classification is available or cannot be determined for the component.",
        examples=["library"],
        title="Component Type",
    )
    mime_type: Optional[constr(regex=r"^[-+a-z0-9.]+/[-+a-z0-9.]+$")] = Field(
        None,
        alias="mime-type",
        description="The optional mime-type of the component. When used on file components, the mime-type can provide additional context about the kind of file being represented such as an image, font, or executable. Some library or framework components may also have an associated mime-type.",
        examples=["image/jpeg"],
        title="Mime-Type",
    )
    bom_ref: Optional[str] = Field(
        None,
        alias="bom-ref",
        description="An optional identifier which can be used to reference the component elsewhere in the BOM. Every bom-ref should be unique.",
        title="BOM Reference",
    )
    supplier: Optional[OrganizationalEntity] = Field(
        None,
        description=" The organization that supplied the component. The supplier may often be the manufacturer, but may also be a distributor or repackager.",
        title="Component Supplier",
    )
    author: Optional[str] = Field(
        None,
        description="The person(s) or organization(s) that authored the component",
        examples=["Acme Inc"],
        title="Component Author",
    )
    publisher: Optional[str] = Field(
        None,
        description="The person(s) or organization(s) that published the component",
        examples=["Acme Inc"],
        title="Component Publisher",
    )
    group: Optional[str] = Field(
        None,
        description="The grouping name or identifier. This will often be a shortened, single name of the company or project that produced the component, or the source package or domain name. Whitespace and special characters should be avoided. Examples include: apache, org.apache.commons, and apache.org.",
        examples=["com.acme"],
        title="Component Group",
    )
    name: str = Field(
        ...,
        description="The name of the component. This will often be a shortened, single name of the component. Examples: commons-lang3 and jquery",
        examples=["tomcat-catalina"],
        title="Component Name",
    )
    version: str = Field(
        ...,
        description="The component version. The version should ideally comply with semantic versioning but is not enforced.",
        examples=["9.0.14"],
        title="Component Version",
    )
    description: Optional[str] = Field(
        None,
        description="Specifies a description for the component",
        title="Component Description",
    )
    scope: Optional[Scope] = Field(
        "required",
        description="Specifies the scope of the component. If scope is not specified, 'required' scope should be assumed by the consumer of the BOM",
        title="Component Scope",
    )
    hashes: Optional[List[Hash]] = Field(None, title="Component Hashes")
    licenses: Optional[List[LicenseChoice]] = Field(None, title="Component License(s)")
    copyright: Optional[str] = Field(
        None,
        description="An optional copyright notice informing users of the underlying claims to copyright ownership in a published work.",
        examples=["Acme Inc"],
        title="Component Copyright",
    )
    cpe: Optional[str] = Field(
        None,
        description="DEPRECATED - DO NOT USE. This will be removed in a future version. Specifies a well-formed CPE name. See https://nvd.nist.gov/products/cpe",
        examples=["cpe:2.3:a:acme:component_framework:-:*:*:*:*:*:*:*"],
        title="Component Common Platform Enumeration (CPE)",
    )
    purl: Optional[str] = Field(
        None,
        examples=["pkg:maven/com.acme/tomcat-catalina@9.0.14?packaging=jar"],
        title="Component Package URL (purl)",
    )
    swid: Optional[Swid] = Field(
        None,
        description="Specifies metadata and content for ISO-IEC 19770-2 Software Identification (SWID) Tags.",
        title="SWID Tag",
    )
    modified: Optional[bool] = Field(
        None,
        description="DEPRECATED - DO NOT USE. This will be removed in a future version. Use the pedigree element instead to supply information on exactly how the component was modified. A boolean value indicating is the component has been modified from the original. A value of true indicates the component is a derivative of the original. A value of false indicates the component has not been modified from the original.",
        title="Component Modified From Original",
    )
    pedigree: Optional[Pedigree] = Field(
        None,
        description="Component pedigree is a way to document complex supply chain scenarios where components are created, distributed, modified, redistributed, combined with other components, etc. Pedigree supports viewing this complex chain from the beginning, the end, or anywhere in the middle. It also provides a way to document variants where the exact relation may not be known.",
        title="Component Pedigree",
    )
    externalReferences: Optional[List[ExternalReference]] = Field(
        None, title="External References"
    )
    components: Optional[List[Component]] = Field(None, title="Components")
    evidence: Optional[ComponentEvidence] = Field(
        None,
        description="Provides the ability to document evidence collected through various forms of extraction or analysis.",
        title="Evidence",
    )
    properties: Optional[List[Property]] = Field(
        None,
        description="Provides the ability to document properties in a name-value store. This provides flexibility to include data not officially supported in the standard without having to use additional namespaces or create extensions. Unlike key-value stores, properties support duplicate names, each potentially having different values.",
        title="Properties",
    )


Service.update_forward_refs()
CyclonedxSoftwareBillOfMaterialSpecification.update_forward_refs()
Metadata.update_forward_refs()
Pedigree.update_forward_refs()



"""
Define context in which a plug-in runs
"""
from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

# Only import these classes during type-checking
#
# If we import at all times, on execution we get an error like
#    "ImportError: cannot import name 'Credentials' from partially initialized module
#         'hoppr.configs.credentials' (most likely due to a circular import)"
#
# If we don't import for type checking, mypy fails.

if TYPE_CHECKING:  # pragma: no cover
    from hoppr.configs.credentials import Credentials
    from hoppr.configs.manifest import Manifest
    from hoppr.types.cyclonedx_1_4 import (
        CyclonedxSoftwareBillOfMaterialsStandard as Bom,
    )


@dataclass
class Context:
    """
    Define context in which a plug-in runs
    """

    manifest: Manifest
    credentials: Credentials
    collect_root_dir: str
    consolidated_sbom: Bom
    max_attempts: int = 3
    retry_wait_seconds: float = 5



# generated by datamodel-codegen:
#   filename:  bom-1.4.schema.json
#   timestamp: 2022-04-27T20:56:27+00:00

from __future__ import annotations

from datetime import datetime
from enum import Enum
from typing import Any, List, Optional, Union

from pydantic import AnyUrl, BaseModel, EmailStr, Extra, Field, conint, constr

from . import spdx


class _Schema(Enum):
    http___cyclonedx_org_schema_bom_1_4_schema_json = (
        "http://cyclonedx.org/schema/bom-1.4.schema.json"
    )


class BomFormat(Enum):
    CycloneDX = "CycloneDX"


class RefType(BaseModel):
    __root__: str


class OrganizationalContact(BaseModel):
    class Config:
        extra = Extra.forbid

    name: Optional[str] = Field(
        None,
        description="The name of a contact",
        examples=["Contact name"],
        title="Name",
    )
    email: Optional[EmailStr] = Field(
        None,
        description="The email address of the contact.",
        examples=["firstname.lastname@example.com"],
        title="Email Address",
    )
    phone: Optional[str] = Field(
        None,
        description="The phone number of the contact.",
        examples=["800-555-1212"],
        title="Phone",
    )


class Type(Enum):
    application = "application"
    framework = "framework"
    library = "library"
    container = "container"
    operating_system = "operating-system"
    device = "device"
    firmware = "firmware"
    file = "file"


class Scope(Enum):
    required = "required"
    optional = "optional"
    excluded = "excluded"


class Encoding(Enum):
    base64 = "base64"


class Attachment(BaseModel):
    class Config:
        extra = Extra.forbid

    contentType: Optional[str] = Field(
        "text/plain",
        description="Specifies the content type of the text. Defaults to text/plain if not specified.",
        title="Content-Type",
    )
    encoding: Optional[Encoding] = Field(
        None,
        description="Specifies the optional encoding the text is represented in.",
        title="Encoding",
    )
    content: str = Field(
        ...,
        description="The attachment data. Proactive controls such as input validation and sanitization should be employed to prevent misuse of attachment text.",
        title="Attachment Text",
    )


class HashAlg(Enum):
    MD5 = "MD5"
    SHA_1 = "SHA-1"
    SHA_256 = "SHA-256"
    SHA_384 = "SHA-384"
    SHA_512 = "SHA-512"
    SHA3_256 = "SHA3-256"
    SHA3_384 = "SHA3-384"
    SHA3_512 = "SHA3-512"
    BLAKE2b_256 = "BLAKE2b-256"
    BLAKE2b_384 = "BLAKE2b-384"
    BLAKE2b_512 = "BLAKE2b-512"
    BLAKE3 = "BLAKE3"


class HashContent(BaseModel):
    __root__: constr(
        regex=r"^([a-fA-F0-9]{32}|[a-fA-F0-9]{40}|[a-fA-F0-9]{64}|[a-fA-F0-9]{96}|[a-fA-F0-9]{128})$"
    ) = Field(
        ..., examples=["3942447fac867ae5cdb3229b658f4d48"], title="Hash Content (value)"
    )


class Type1(Enum):
    unofficial = "unofficial"
    monkey = "monkey"
    backport = "backport"
    cherry_pick = "cherry-pick"


class Diff(BaseModel):
    class Config:
        extra = Extra.forbid

    text: Optional[Attachment] = Field(
        None, description="Specifies the optional text of the diff", title="Diff text"
    )
    url: Optional[str] = Field(
        None, description="Specifies the URL to the diff", title="URL"
    )


class Type2(Enum):
    defect = "defect"
    enhancement = "enhancement"
    security = "security"


class Source(BaseModel):
    class Config:
        extra = Extra.forbid

    name: Optional[str] = Field(
        None,
        description="The name of the source. For example 'National Vulnerability Database', 'NVD', and 'Apache'",
        title="Name",
    )
    url: Optional[str] = Field(
        None,
        description="The url of the issue documentation as provided by the source",
        title="URL",
    )


class Issue(BaseModel):
    class Config:
        extra = Extra.forbid

    type: Type2 = Field(..., description="Specifies the type of issue", title="Type")
    id: Optional[str] = Field(
        None,
        description="The identifier of the issue assigned by the source of the issue",
        title="ID",
    )
    name: Optional[str] = Field(None, description="The name of the issue", title="Name")
    description: Optional[str] = Field(
        None, description="A description of the issue", title="Description"
    )
    source: Optional[Source] = Field(
        None,
        description="The source of the issue where it is documented",
        title="Source",
    )
    references: Optional[List[str]] = Field(
        None,
        description="A collection of URL's for reference. Multiple URLs are allowed.",
        examples=["https://example.com"],
        title="References",
    )


class IdentifiableAction(BaseModel):
    class Config:
        extra = Extra.forbid

    timestamp: Optional[datetime] = Field(
        None,
        description="The timestamp in which the action occurred",
        title="Timestamp",
    )
    name: Optional[str] = Field(
        None,
        description="The name of the individual who performed the action",
        title="Name",
    )
    email: Optional[EmailStr] = Field(
        None,
        description="The email address of the individual who performed the action",
        title="E-mail",
    )


class Type3(Enum):
    vcs = "vcs"
    issue_tracker = "issue-tracker"
    website = "website"
    advisories = "advisories"
    bom = "bom"
    mailing_list = "mailing-list"
    social = "social"
    chat = "chat"
    documentation = "documentation"
    support = "support"
    distribution = "distribution"
    license = "license"
    build_meta = "build-meta"
    build_system = "build-system"
    release_notes = "release-notes"
    other = "other"


class Dependency(BaseModel):
    class Config:
        extra = Extra.forbid

    ref: RefType = Field(
        ...,
        description="References a component by the components bom-ref attribute",
        title="Reference",
    )
    dependsOn: Optional[List[RefType]] = Field(
        None,
        description="The bom-ref identifiers of the components that are dependencies of this dependency object.",
        title="Depends On",
        # REMOVED IDH unique_items=True,
    )


class DataFlow(Enum):
    inbound = "inbound"
    outbound = "outbound"
    bi_directional = "bi-directional"
    unknown = "unknown"


class Copyright(BaseModel):
    class Config:
        extra = Extra.forbid

    text: str = Field(..., title="Copyright Text")


class AggregateType(Enum):
    complete = "complete"
    incomplete = "incomplete"
    incomplete_first_party_only = "incomplete_first_party_only"
    incomplete_third_party_only = "incomplete_third_party_only"
    unknown = "unknown"
    not_specified = "not_specified"


class Property(BaseModel):
    name: Optional[str] = Field(
        None,
        description="The name of the property. Duplicate names are allowed, each potentially having a different value.",
        title="Name",
    )
    value: Optional[str] = Field(
        None, description="The value of the property.", title="Value"
    )


class LocaleType(BaseModel):
    __root__: constr(regex=r"^([a-z]{2})(-[A-Z]{2})?$") = Field(
        ...,
        description="Defines a syntax for representing two character language code (ISO-639) followed by an optional two character country code. The language code MUST be lower case. If the country code is specified, the country code MUST be upper case. The language code and country code MUST be separated by a minus sign. Examples: en, en-US, fr, fr-CA",
        title="Locale",
    )


class ReleaseType(BaseModel):
    __root__: str = Field(
        ...,
        description="The software versioning type. It is RECOMMENDED that the release type use one of 'major', 'minor', 'patch', 'pre-release', or 'internal'. Representing all possible software release types is not practical, so standardizing on the recommended values, whenever possible, is strongly encouraged.\n\n* __major__ = A major release may contain significant changes or may introduce breaking changes.\n* __minor__ = A minor release, also known as an update, may contain a smaller number of changes than major releases.\n* __patch__ = Patch releases are typically unplanned and may resolve defects or important security issues.\n* __pre-release__ = A pre-release may include alpha, beta, or release candidates and typically have limited support. They provide the ability to preview a release prior to its general availability.\n* __internal__ = Internal releases are not for public consumption and are intended to be used exclusively by the project or manufacturer that produced it.",
        examples=["major", "minor", "patch", "pre-release", "internal"],
    )


class Note(BaseModel):
    class Config:
        extra = Extra.forbid

    locale: Optional[LocaleType] = Field(
        None,
        description='The ISO-639 (or higher) language code and optional ISO-3166 (or higher) country code. Examples include: "en", "en-US", "fr" and "fr-CA"',
        title="Locale",
    )
    text: Attachment = Field(
        ...,
        description="Specifies the full content of the release note.",
        title="Release note content",
    )


class ReleaseNotes(BaseModel):
    class Config:
        extra = Extra.forbid

    type: ReleaseType = Field(
        ...,
        description="The software versioning type the release note describes.",
        title="Type",
    )
    title: Optional[str] = Field(
        None, description="The title of the release.", title="Title"
    )
    featuredImage: Optional[str] = Field(
        None,
        description="The URL to an image that may be prominently displayed with the release note.",
        title="Featured image",
    )
    socialImage: Optional[str] = Field(
        None,
        description="The URL to an image that may be used in messaging on social media platforms.",
        title="Social image",
    )
    description: Optional[str] = Field(
        None, description="A short description of the release.", title="Description"
    )
    timestamp: Optional[datetime] = Field(
        None,
        description="The date and time (timestamp) when the release note was created.",
        title="Timestamp",
    )
    aliases: Optional[List[str]] = Field(
        None,
        description="One or more alternate names the release may be referred to. This may include unofficial terms used by development and marketing teams (e.g. code names).",
        title="Aliases",
    )
    tags: Optional[List[str]] = Field(
        None,
        description="One or more tags that may aid in search or retrieval of the release note.",
        title="Tags",
    )
    resolves: Optional[List[Issue]] = Field(
        None,
        description="A collection of issues that have been resolved.",
        title="Resolves",
    )
    notes: Optional[List[Note]] = Field(
        None,
        description="Zero or more release notes containing the locale and content. Multiple note objects may be specified to support release notes in a wide variety of languages.",
        title="Notes",
    )
    properties: Optional[List[Property]] = Field(
        None,
        description="Provides the ability to document properties in a name-value store. This provides flexibility to include data not officially supported in the standard without having to use additional namespaces or create extensions. Unlike key-value stores, properties support duplicate names, each potentially having different values. Property names of interest to the general public are encouraged to be registered in the [CycloneDX Property Taxonomy](https://github.com/CycloneDX/cyclonedx-property-taxonomy). Formal registration is OPTIONAL.",
        title="Properties",
    )


class Advisory(BaseModel):
    class Config:
        extra = Extra.forbid

    title: Optional[str] = Field(
        None, description="An optional name of the advisory.", title="Title"
    )
    url: str = Field(
        ..., description="Location where the advisory can be obtained.", title="URL"
    )


class Cwe(BaseModel):
    __root__: conint(ge=1) = Field(
        ...,
        description="Integer representation of a Common Weaknesses Enumerations (CWE). For example 399 (of https://cwe.mitre.org/data/definitions/399.html)",
        title="CWE",
    )


class Severity(Enum):
    critical = "critical"
    high = "high"
    medium = "medium"
    low = "low"
    info = "info"
    none = "none"
    unknown = "unknown"


class ScoreMethod(Enum):
    CVSSv2 = "CVSSv2"
    CVSSv3 = "CVSSv3"
    CVSSv31 = "CVSSv31"
    OWASP = "OWASP"
    other = "other"


class ImpactAnalysisState(Enum):
    resolved = "resolved"
    resolved_with_pedigree = "resolved_with_pedigree"
    exploitable = "exploitable"
    in_triage = "in_triage"
    false_positive = "false_positive"
    not_affected = "not_affected"


class ImpactAnalysisJustification(Enum):
    code_not_present = "code_not_present"
    code_not_reachable = "code_not_reachable"
    requires_configuration = "requires_configuration"
    requires_dependency = "requires_dependency"
    requires_environment = "requires_environment"
    protected_by_compiler = "protected_by_compiler"
    protected_at_runtime = "protected_at_runtime"
    protected_at_perimeter = "protected_at_perimeter"
    protected_by_mitigating_control = "protected_by_mitigating_control"


class VulnerabilitySource(BaseModel):
    class Config:
        extra = Extra.forbid

    url: Optional[str] = Field(
        None,
        description="The url of the vulnerability documentation as provided by the source.",
        examples=["https://nvd.nist.gov/vuln/detail/CVE-2021-39182"],
        title="URL",
    )
    name: Optional[str] = Field(
        None,
        description="The name of the source.",
        examples=[
            "NVD",
            "National Vulnerability Database",
            "OSS Index",
            "VulnDB",
            "GitHub Advisories",
        ],
        title="Name",
    )


class Reference(BaseModel):
    class Config:
        extra = Extra.forbid

    id: str = Field(
        ...,
        description="An identifier that uniquely identifies the vulnerability.",
        examples=[
            "CVE-2021-39182",
            "GHSA-35m5-8cvj-8783",
            "SNYK-PYTHON-ENROCRYPT-1912876",
        ],
        title="ID",
    )
    source: VulnerabilitySource = Field(
        ..., description="The source that published the vulnerability."
    )


class ResponseEnum(Enum):
    can_not_fix = "can_not_fix"
    will_not_fix = "will_not_fix"
    update = "update"
    rollback = "rollback"
    workaround_available = "workaround_available"


class Analysis(BaseModel):
    class Config:
        extra = Extra.forbid

    state: Optional[ImpactAnalysisState] = None
    justification: Optional[ImpactAnalysisJustification] = None
    response: Optional[List[ResponseEnum]] = Field(
        None,
        description="A response to the vulnerability by the manufacturer, supplier, or project responsible for the affected component or service. More than one response is allowed. Responses are strongly encouraged for vulnerabilities where the analysis state is exploitable.",
        title="Response",
    )
    detail: Optional[str] = Field(
        None,
        description="Detailed description of the impact including methods used during assessment. If a vulnerability is not exploitable, this field should include specific details on why the component or service is not impacted by this vulnerability.",
        title="Detail",
    )


class Affect(BaseModel):
    class Config:
        extra = Extra.forbid

    ref: RefType = Field(
        ...,
        description="References a component or service by the objects bom-ref",
        title="Reference",
    )
    versions: Optional[List[Union[Any, Any]]] = Field(
        None,
        description="Zero or more individual versions or range of versions.",
        title="Versions",
    )


class AffectedStatus(Enum):
    affected = "affected"
    unaffected = "unaffected"
    unknown = "unknown"


class Version(BaseModel):
    __root__: constr(min_length=1, max_length=1024) = Field(
        ..., description="A single version of a component or service."
    )


class Range(BaseModel):
    __root__: constr(min_length=1, max_length=1024) = Field(
        ...,
        description="A version range specified in Package URL Version Range syntax (vers) which is defined at https://github.com/package-url/purl-spec/VERSION-RANGE-SPEC.rst",
    )


class AlgorithmEnum(Enum):
    RS256 = "RS256"
    RS384 = "RS384"
    RS512 = "RS512"
    PS256 = "PS256"
    PS384 = "PS384"
    PS512 = "PS512"
    ES256 = "ES256"
    ES384 = "ES384"
    ES512 = "ES512"
    Ed25519 = "Ed25519"
    Ed448 = "Ed448"
    HS256 = "HS256"
    HS384 = "HS384"
    HS512 = "HS512"


class KeyType(Enum):
    EC = "EC"
    OKP = "OKP"
    RSA = "RSA"


class PublicKey(BaseModel):
    class Config:
        extra = Extra.allow

    kty: KeyType


class OrganizationalEntity(BaseModel):
    class Config:
        extra = Extra.forbid

    name: Optional[str] = Field(
        None,
        description="The name of the organization",
        examples=["Example Inc."],
        title="Name",
    )
    url: Optional[List[str]] = Field(
        None,
        description="The URL of the organization. Multiple URLs are allowed.",
        examples=["https://example.com"],
        title="URL",
    )
    contact: Optional[List[OrganizationalContact]] = Field(
        None,
        description="A contact at the organization. Multiple contacts are allowed.",
        title="Contact",
    )


class Swid(BaseModel):
    class Config:
        extra = Extra.forbid

    tagId: str = Field(
        ..., description="Maps to the tagId of a SoftwareIdentity.", title="Tag ID"
    )
    name: str = Field(
        ..., description="Maps to the name of a SoftwareIdentity.", title="Name"
    )
    version: Optional[str] = Field(
        "0.0", description="Maps to the version of a SoftwareIdentity.", title="Version"
    )
    tagVersion: Optional[int] = Field(
        0,
        description="Maps to the tagVersion of a SoftwareIdentity.",
        title="Tag Version",
    )
    patch: Optional[bool] = Field(
        False, description="Maps to the patch of a SoftwareIdentity.", title="Patch"
    )
    text: Optional[Attachment] = Field(
        None,
        description="Specifies the metadata and content of the SWID tag.",
        title="Attachment text",
    )
    url: Optional[str] = Field(
        None, description="The URL to the SWID file.", title="URL"
    )


class Hash(BaseModel):
    class Config:
        extra = Extra.forbid

    alg: HashAlg
    content: HashContent


class License(BaseModel):
    class Config:
        extra = Extra.forbid

    id: Optional[spdx.Schema] = Field(
        None,
        description="A valid SPDX license ID",
        examples=["Apache-2.0"],
        title="License ID (SPDX)",
    )
    name: Optional[str] = Field(
        None,
        description="If SPDX does not define the license used, this field may be used to provide the license name",
        examples=["Acme Software License"],
        title="License Name",
    )
    text: Optional[Attachment] = Field(
        None,
        description="An optional way to include the textual content of a license.",
        title="License text",
    )
    url: Optional[str] = Field(
        None,
        description="The URL to the license file. If specified, a 'license' externalReference should also be specified for completeness",
        examples=["https://www.apache.org/licenses/LICENSE-2.0.txt"],
        title="License URL",
    )


class License1(BaseModel):
    __root__: Union[License, Any, Any] = Field(..., title="License Object")


class LicenseChoice1(BaseModel):
    class Config:
        extra = Extra.forbid

    license: Optional[License1] = None
    expression: Optional[str] = Field(
        None,
        examples=[
            "Apache-2.0 AND (MIT OR GPL-2.0-only)",
            "GPL-3.0-only WITH Classpath-exception-2.0",
        ],
        title="SPDX License Expression",
    )


class LicenseChoice(BaseModel):
    __root__: Union[LicenseChoice1, Any, Any] = Field(..., title="License(s)")


class Commit(BaseModel):
    class Config:
        extra = Extra.forbid

    uid: Optional[str] = Field(
        None,
        description="A unique identifier of the commit. This may be version control specific. For example, Subversion uses revision numbers whereas git uses commit hashes.",
        title="UID",
    )
    url: Optional[str] = Field(
        None,
        description="The URL to the commit. This URL will typically point to a commit in a version control system.",
        title="URL",
    )
    author: Optional[IdentifiableAction] = Field(
        None,
        description="The author who created the changes in the commit",
        title="Author",
    )
    committer: Optional[IdentifiableAction] = Field(
        None,
        description="The person who committed or pushed the commit",
        title="Committer",
    )
    message: Optional[str] = Field(
        None,
        description="The text description of the contents of the commit",
        title="Message",
    )


class Patch(BaseModel):
    class Config:
        extra = Extra.forbid

    type: Type1 = Field(
        ...,
        description="Specifies the purpose for the patch including the resolution of defects, security issues, or new behavior or functionality.\n\n* __unofficial__ = A patch which is not developed by the creators or maintainers of the software being patched. Refer to [https://en.wikipedia.org/wiki/Unofficial_patch](https://en.wikipedia.org/wiki/Unofficial_patch)\n* __monkey__ = A patch which dynamically modifies runtime behavior. Refer to [https://en.wikipedia.org/wiki/Monkey_patch](https://en.wikipedia.org/wiki/Monkey_patch)\n* __backport__ = A patch which takes code from a newer version of software and applies it to older versions of the same software. Refer to [https://en.wikipedia.org/wiki/Backporting](https://en.wikipedia.org/wiki/Backporting)\n* __cherry-pick__ = A patch created by selectively applying commits from other versions or branches of the same software.",
        title="Type",
    )
    diff: Optional[Diff] = Field(
        None,
        description="The patch file (or diff) that show changes. Refer to [https://en.wikipedia.org/wiki/Diff](https://en.wikipedia.org/wiki/Diff)",
        title="Diff",
    )
    resolves: Optional[List[Issue]] = Field(
        None, description="A collection of issues the patch resolves", title="Resolves"
    )


class ExternalReference(BaseModel):
    class Config:
        extra = Extra.forbid

    url: str = Field(..., description="The URL to the external reference", title="URL")
    comment: Optional[str] = Field(
        None,
        description="An optional comment describing the external reference",
        title="Comment",
    )
    type: Type3 = Field(
        ...,
        description='Specifies the type of external reference. There are built-in types to describe common references. If a type does not exist for the reference being referred to, use the "other" type.',
        title="Type",
    )
    hashes: Optional[List[Hash]] = Field(
        None,
        description="The hashes of the external reference (if applicable).",
        title="Hashes",
    )


class DataClassification(BaseModel):
    class Config:
        extra = Extra.forbid

    flow: DataFlow = Field(
        ...,
        description="Specifies the flow direction of the data. Direction is relative to the service. Inbound flow states that data enters the service. Outbound flow states that data leaves the service. Bi-directional states that data flows both ways, and unknown states that the direction is not known.",
        title="Directional Flow",
    )
    classification: str = Field(
        ...,
        description="Data classification tags data according to its type, sensitivity, and value if altered, stolen, or destroyed.",
        title="Classification",
    )


class ComponentEvidence(BaseModel):
    class Config:
        extra = Extra.forbid

    licenses: Optional[List[LicenseChoice]] = Field(None, title="Component License(s)")
    copyright: Optional[List[Copyright]] = Field(None, title="Copyright")


class Rating(BaseModel):
    class Config:
        extra = Extra.forbid

    source: Optional[VulnerabilitySource] = Field(
        None,
        description="The source that calculated the severity or risk rating of the vulnerability.",
    )
    score: Optional[float] = Field(
        None, description="The numerical score of the rating.", title="Score"
    )
    severity: Optional[Severity] = Field(
        None,
        description="Textual representation of the severity that corresponds to the numerical score of the rating.",
    )
    method: Optional[ScoreMethod] = None
    vector: Optional[str] = Field(
        None,
        description="Textual representation of the metric values used to score the vulnerability",
        title="Vector",
    )
    justification: Optional[str] = Field(
        None,
        description="An optional reason for rating the vulnerability as it was",
        title="Justification",
    )


class Credits(BaseModel):
    class Config:
        extra = Extra.forbid

    organizations: Optional[List[OrganizationalEntity]] = Field(
        None,
        description="The organizations credited with vulnerability discovery.",
        title="Organizations",
    )
    individuals: Optional[List[OrganizationalContact]] = Field(
        None,
        description="The individuals, not associated with organizations, that are credited with vulnerability discovery.",
        title="Individuals",
    )


class Signer(BaseModel):
    class Config:
        extra = Extra.forbid

    algorithm: Union[AlgorithmEnum, AnyUrl]
    keyId: Optional[str] = Field(
        None,
        description="Optional. Application specific string identifying the signature key.",
        title="Key ID",
    )
    publicKey: Optional[PublicKey] = Field(
        None, description="Optional. Public key object.", title="Public key"
    )
    certificatePath: Optional[List[str]] = Field(
        None,
        description="Optional. Sorted array of X.509 [RFC5280] certificates, where the first element must contain the signature certificate. The certificate path must be contiguous but is not required to be complete.",
        title="Certificate path",
    )
    excludes: Optional[List[str]] = Field(
        None,
        description='Optional. Array holding the names of one or more application level properties that must be excluded from the signature process. Note that the "excludes" property itself, must also be excluded from the signature process. Since both the "excludes" property and the associated data it points to are unsigned, a conforming JSF implementation must provide options for specifying which properties to accept.',
        title="Excludes",
    )
    value: str = Field(
        ...,
        description="The signature data. Note that the binary representation must follow the JWA [RFC7518] specifications.",
        title="Signature",
    )


class Tool(BaseModel):
    class Config:
        extra = Extra.forbid

    vendor: Optional[str] = Field(
        None,
        description="The name of the vendor who created the tool",
        title="Tool Vendor",
    )
    name: Optional[str] = Field(
        None, description="The name of the tool", title="Tool Name"
    )
    version: Optional[str] = Field(
        None, description="The version of the tool", title="Tool Version"
    )
    hashes: Optional[List[Hash]] = Field(
        None, description="The hashes of the tool (if applicable).", title="Hashes"
    )
    externalReferences: Optional[List[ExternalReference]] = Field(
        None,
        description="External references provide a way to document systems, sites, and information that may be relevant but which are not included with the BOM.",
        title="External References",
    )


class Vulnerability(BaseModel):
    class Config:
        extra = Extra.forbid

    bom_ref: Optional[RefType] = Field(
        None,
        alias="bom-ref",
        description="An optional identifier which can be used to reference the vulnerability elsewhere in the BOM. Every bom-ref MUST be unique within the BOM.",
        title="BOM Reference",
    )
    id: Optional[str] = Field(
        None,
        description="The identifier that uniquely identifies the vulnerability.",
        examples=[
            "CVE-2021-39182",
            "GHSA-35m5-8cvj-8783",
            "SNYK-PYTHON-ENROCRYPT-1912876",
        ],
        title="ID",
    )
    source: Optional[VulnerabilitySource] = Field(
        None, description="The source that published the vulnerability."
    )
    references: Optional[List[Reference]] = Field(
        None,
        description="Zero or more pointers to vulnerabilities that are the equivalent of the vulnerability specified. Often times, the same vulnerability may exist in multiple sources of vulnerability intelligence, but have different identifiers. References provide a way to correlate vulnerabilities across multiple sources of vulnerability intelligence.",
        title="References",
    )
    ratings: Optional[List[Rating]] = Field(
        None, description="List of vulnerability ratings", title="Ratings"
    )
    cwes: Optional[List[Cwe]] = Field(
        None,
        description="List of Common Weaknesses Enumerations (CWEs) codes that describes this vulnerability. For example 399 (of https://cwe.mitre.org/data/definitions/399.html)",
        examples=["399"],
        title="CWEs",
    )
    description: Optional[str] = Field(
        None,
        description="A description of the vulnerability as provided by the source.",
        title="Description",
    )
    detail: Optional[str] = Field(
        None,
        description="If available, an in-depth description of the vulnerability as provided by the source organization. Details often include examples, proof-of-concepts, and other information useful in understanding root cause.",
        title="Details",
    )
    recommendation: Optional[str] = Field(
        None,
        description="Recommendations of how the vulnerability can be remediated or mitigated.",
        title="Details",
    )
    advisories: Optional[List[Advisory]] = Field(
        None,
        description="Published advisories of the vulnerability if provided.",
        title="Advisories",
    )
    created: Optional[datetime] = Field(
        None,
        description="The date and time (timestamp) when the vulnerability record was created in the vulnerability database.",
        title="Created",
    )
    published: Optional[datetime] = Field(
        None,
        description="The date and time (timestamp) when the vulnerability record was first published.",
        title="Published",
    )
    updated: Optional[datetime] = Field(
        None,
        description="The date and time (timestamp) when the vulnerability record was last updated.",
        title="Updated",
    )
    credits: Optional[Credits] = Field(
        None,
        description="Individuals or organizations credited with the discovery of the vulnerability.",
        title="Credits",
    )
    tools: Optional[List[Tool]] = Field(
        None,
        description="The tool(s) used to identify, confirm, or score the vulnerability.",
        title="Creation Tools",
    )
    analysis: Optional[Analysis] = Field(
        None,
        description="An assessment of the impact and exploitability of the vulnerability.",
        title="Impact Analysis",
    )
    affects: Optional[List[Affect]] = Field(
        None,
        description="The components or services that are affected by the vulnerability.",
        title="Affects",
        # REMOVED IDH unique_items=True,
    )
    properties: Optional[List[Property]] = Field(
        None,
        description="Provides the ability to document properties in a name-value store. This provides flexibility to include data not officially supported in the standard without having to use additional namespaces or create extensions. Unlike key-value stores, properties support duplicate names, each potentially having different values. Property names of interest to the general public are encouraged to be registered in the [CycloneDX Property Taxonomy](https://github.com/CycloneDX/cyclonedx-property-taxonomy). Formal registration is OPTIONAL.",
        title="Properties",
    )


class SignatureItem(BaseModel):
    class Config:
        extra = Extra.forbid

    signers: Optional[List[Signer]] = Field(
        None,
        description="Unique top level property for Multiple Signatures. (multisignature)",
        title="Signature",
    )


class SignatureItem1(BaseModel):
    class Config:
        extra = Extra.forbid

    chain: Optional[List[Signer]] = Field(
        None,
        description="Unique top level property for Signature Chains. (signaturechain)",
        title="Signature",
    )


class Signature(BaseModel):
    __root__: Union[SignatureItem, SignatureItem1, Signer] = Field(
        ..., title="Signature"
    )


class Service(BaseModel):
    class Config:
        extra = Extra.forbid

    bom_ref: Optional[RefType] = Field(
        None,
        alias="bom-ref",
        description="An optional identifier which can be used to reference the service elsewhere in the BOM. Every bom-ref MUST be unique within the BOM.",
        title="BOM Reference",
    )
    provider: Optional[OrganizationalEntity] = Field(
        None,
        description="The organization that provides the service.",
        title="Provider",
    )
    group: Optional[str] = Field(
        None,
        description="The grouping name, namespace, or identifier. This will often be a shortened, single name of the company or project that produced the service or domain name. Whitespace and special characters should be avoided.",
        examples=["com.acme"],
        title="Service Group",
    )
    name: str = Field(
        ...,
        description="The name of the service. This will often be a shortened, single name of the service.",
        examples=["ticker-service"],
        title="Service Name",
    )
    version: Optional[str] = Field(
        None,
        description="The service version.",
        examples=["1.0.0"],
        title="Service Version",
    )
    description: Optional[str] = Field(
        None,
        description="Specifies a description for the service",
        title="Service Description",
    )
    endpoints: Optional[List[str]] = Field(
        None,
        description="The endpoint URIs of the service. Multiple endpoints are allowed.",
        examples=["https://example.com/api/v1/ticker"],
        title="Endpoints",
    )
    authenticated: Optional[bool] = Field(
        None,
        description="A boolean value indicating if the service requires authentication. A value of true indicates the service requires authentication prior to use. A value of false indicates the service does not require authentication.",
        title="Authentication Required",
    )
    x_trust_boundary: Optional[bool] = Field(
        None,
        alias="x-trust-boundary",
        description="A boolean value indicating if use of the service crosses a trust zone or boundary. A value of true indicates that by using the service, a trust boundary is crossed. A value of false indicates that by using the service, a trust boundary is not crossed.",
        title="Crosses Trust Boundary",
    )
    data: Optional[List[DataClassification]] = Field(
        None,
        description="Specifies the data classification.",
        title="Data Classification",
    )
    licenses: Optional[List[LicenseChoice]] = Field(None, title="Component License(s)")
    externalReferences: Optional[List[ExternalReference]] = Field(
        None,
        description="External references provide a way to document systems, sites, and information that may be relevant but which are not included with the BOM.",
        title="External References",
    )
    services: Optional[List[Service]] = Field(
        None,
        description="A list of services included or deployed behind the parent service. This is not a dependency tree. It provides a way to specify a hierarchical representation of service assemblies.",
        title="Services",
    )
    releaseNotes: Optional[ReleaseNotes] = Field(
        None, description="Specifies optional release notes.", title="Release notes"
    )
    properties: Optional[List[Property]] = Field(
        None,
        description="Provides the ability to document properties in a name-value store. This provides flexibility to include data not officially supported in the standard without having to use additional namespaces or create extensions. Unlike key-value stores, properties support duplicate names, each potentially having different values. Property names of interest to the general public are encouraged to be registered in the [CycloneDX Property Taxonomy](https://github.com/CycloneDX/cyclonedx-property-taxonomy). Formal registration is OPTIONAL.",
        title="Properties",
    )
    signature: Optional[Signature] = Field(
        None,
        description="Enveloped signature in [JSON Signature Format (JSF)](https://cyberphone.github.io/doc/security/jsf.html).",
        title="Signature",
    )


class Compositions(BaseModel):
    class Config:
        extra = Extra.forbid

    aggregate: AggregateType = Field(
        ...,
        description="Specifies an aggregate type that describe how complete a relationship is.",
        title="Aggregate",
    )
    assemblies: Optional[List[str]] = Field(
        None,
        description="The bom-ref identifiers of the components or services being described. Assemblies refer to nested relationships whereby a constituent part may include other constituent parts. References do not cascade to child parts. References are explicit for the specified constituent part only.",
        title="BOM references",
        # REMOVED IDH unique_items=True,
    )
    dependencies: Optional[List[str]] = Field(
        None,
        description="The bom-ref identifiers of the components or services being described. Dependencies refer to a relationship whereby an independent constituent part requires another independent constituent part. References do not cascade to transitive dependencies. References are explicit for the specified dependency only.",
        title="BOM references",
        # REMOVED IDH unique_items=True,
    )
    signature: Optional[Signature] = Field(
        None,
        description="Enveloped signature in [JSON Signature Format (JSF)](https://cyberphone.github.io/doc/security/jsf.html).",
        title="Signature",
    )


class CyclonedxSoftwareBillOfMaterialsStandard(BaseModel):
    class Config:
        # REMOVED IDH extra = Extra.forbid
        pass

    _schema: Optional[_Schema] = Field(None, alias="$schema")
    bomFormat: BomFormat = Field(
        ...,
        description='Specifies the format of the BOM. This helps to identify the file as CycloneDX since BOMs do not have a filename convention nor does JSON schema support namespaces. This value MUST be "CycloneDX".',
        title="BOM Format",
    )
    specVersion: str = Field(
        ...,
        description="The version of the CycloneDX specification a BOM conforms to (starting at version 1.2).",
        examples=["1.4"],
        title="CycloneDX Specification Version",
    )
    serialNumber: Optional[
        constr(
            regex=r"^urn:uuid:[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$"
        )
    ] = Field(
        None,
        description="Every BOM generated SHOULD have a unique serial number, even if the contents of the BOM have not changed over time. If specified, the serial number MUST conform to RFC-4122. Use of serial numbers are RECOMMENDED.",
        examples=["urn:uuid:3e671687-395b-41f5-a30f-a58921a69b79"],
        title="BOM Serial Number",
    )
    version: int = Field(
        ...,
        description="Whenever an existing BOM is modified, either manually or through automated processes, the version of the BOM SHOULD be incremented by 1. When a system is presented with multiple BOMs with identical serial numbers, the system SHOULD use the most recent version of the BOM. The default version is '1'.",
        examples=[1],
        title="BOM Version",
    )
    metadata: Optional[Metadata] = Field(
        None,
        description="Provides additional information about a BOM.",
        title="BOM Metadata",
    )
    components: Optional[List[Component]] = Field(
        None,
        description="A list of software and hardware components.",
        title="Components",
        # REMOVED IDH unique_items=True,
    )
    services: Optional[List[Service]] = Field(
        None,
        description="A list of services. This may include microservices, function-as-a-service, and other types of network or intra-process services.",
        title="Services",
        # REMOVED IDH unique_items=True,
    )
    externalReferences: Optional[List[ExternalReference]] = Field(
        None,
        description="External references provide a way to document systems, sites, and information that may be relevant but which are not included with the BOM.",
        title="External References",
    )
    dependencies: Optional[List[Dependency]] = Field(
        None,
        description="Provides the ability to document dependency relationships.",
        title="Dependencies",
        # REMOVED IDH unique_items=True,
    )
    compositions: Optional[List[Compositions]] = Field(
        None,
        description="Compositions describe constituent parts (including components, services, and dependency relationships) and their completeness.",
        title="Compositions",
        # REMOVED IDH unique_items=True,
    )
    vulnerabilities: Optional[List[Vulnerability]] = Field(
        None,
        description="Vulnerabilities identified in components or services.",
        title="Vulnerabilities",
        # REMOVED IDH unique_items=True,
    )
    signature: Optional[Signature] = Field(
        None,
        description="Enveloped signature in [JSON Signature Format (JSF)](https://cyberphone.github.io/doc/security/jsf.html).",
        title="Signature",
    )


class Metadata(BaseModel):
    class Config:
        extra = Extra.forbid

    timestamp: Optional[datetime] = Field(
        None,
        description="The date and time (timestamp) when the BOM was created.",
        title="Timestamp",
    )
    tools: Optional[List[Tool]] = Field(
        None,
        description="The tool(s) used in the creation of the BOM.",
        title="Creation Tools",
    )
    authors: Optional[List[OrganizationalContact]] = Field(
        None,
        description="The person(s) who created the BOM. Authors are common in BOMs created through manual processes. BOMs created through automated means may not have authors.",
        title="Authors",
    )
    component: Optional[Component] = Field(
        None, description="The component that the BOM describes.", title="Component"
    )
    manufacture: Optional[OrganizationalEntity] = Field(
        None,
        description="The organization that manufactured the component that the BOM describes.",
        title="Manufacture",
    )
    supplier: Optional[OrganizationalEntity] = Field(
        None,
        description=" The organization that supplied the component that the BOM describes. The supplier may often be the manufacturer, but may also be a distributor or repackager.",
        title="Supplier",
    )
    licenses: Optional[List[LicenseChoice]] = Field(None, title="BOM License(s)")
    properties: Optional[List[Property]] = Field(
        None,
        description="Provides the ability to document properties in a name-value store. This provides flexibility to include data not officially supported in the standard without having to use additional namespaces or create extensions. Unlike key-value stores, properties support duplicate names, each potentially having different values. Property names of interest to the general public are encouraged to be registered in the [CycloneDX Property Taxonomy](https://github.com/CycloneDX/cyclonedx-property-taxonomy). Formal registration is OPTIONAL.",
        title="Properties",
    )


class Pedigree(BaseModel):
    class Config:
        extra = Extra.forbid

    ancestors: Optional[List[Component]] = Field(
        None,
        description="Describes zero or more components in which a component is derived from. This is commonly used to describe forks from existing projects where the forked version contains a ancestor node containing the original component it was forked from. For example, Component A is the original component. Component B is the component being used and documented in the BOM. However, Component B contains a pedigree node with a single ancestor documenting Component A - the original component from which Component B is derived from.",
        title="Ancestors",
    )
    descendants: Optional[List[Component]] = Field(
        None,
        description="Descendants are the exact opposite of ancestors. This provides a way to document all forks (and their forks) of an original or root component.",
        title="Descendants",
    )
    variants: Optional[List[Component]] = Field(
        None,
        description="Variants describe relations where the relationship between the components are not known. For example, if Component A contains nearly identical code to Component B. They are both related, but it is unclear if one is derived from the other, or if they share a common ancestor.",
        title="Variants",
    )
    commits: Optional[List[Commit]] = Field(
        None,
        description="A list of zero or more commits which provide a trail describing how the component deviates from an ancestor, descendant, or variant.",
        title="Commits",
    )
    patches: Optional[List[Patch]] = Field(
        None,
        description=">A list of zero or more patches describing how the component deviates from an ancestor, descendant, or variant. Patches may be complimentary to commits or may be used in place of commits.",
        title="Patches",
    )
    notes: Optional[str] = Field(
        None,
        description="Notes, observations, and other non-structured commentary describing the components pedigree.",
        title="Notes",
    )


class Component(BaseModel):
    class Config:
        extra = Extra.forbid

    type: Type = Field(
        ...,
        description="Specifies the type of component. For software components, classify as application if no more specific appropriate classification is available or cannot be determined for the component. Types include:\n\n* __application__ = A software application. Refer to [https://en.wikipedia.org/wiki/Application_software](https://en.wikipedia.org/wiki/Application_software) for information about applications.\n* __framework__ = A software framework. Refer to [https://en.wikipedia.org/wiki/Software_framework](https://en.wikipedia.org/wiki/Software_framework) for information on how frameworks vary slightly from libraries.\n* __library__ = A software library. Refer to [https://en.wikipedia.org/wiki/Library_(computing)](https://en.wikipedia.org/wiki/Library_(computing))\n for information about libraries. All third-party and open source reusable components will likely be a library. If the library also has key features of a framework, then it should be classified as a framework. If not, or is unknown, then specifying library is RECOMMENDED.\n* __container__ = A packaging and/or runtime format, not specific to any particular technology, which isolates software inside the container from software outside of a container through virtualization technology. Refer to [https://en.wikipedia.org/wiki/OS-level_virtualization](https://en.wikipedia.org/wiki/OS-level_virtualization)\n* __operating-system__ = A software operating system without regard to deployment model (i.e. installed on physical hardware, virtual machine, image, etc) Refer to [https://en.wikipedia.org/wiki/Operating_system](https://en.wikipedia.org/wiki/Operating_system)\n* __device__ = A hardware device such as a processor, or chip-set. A hardware device containing firmware SHOULD include a component for the physical hardware itself, and another component of type 'firmware' or 'operating-system' (whichever is relevant), describing information about the software running on the device.\n* __firmware__ = A special type of software that provides low-level control over a devices hardware. Refer to [https://en.wikipedia.org/wiki/Firmware](https://en.wikipedia.org/wiki/Firmware)\n* __file__ = A computer file. Refer to [https://en.wikipedia.org/wiki/Computer_file](https://en.wikipedia.org/wiki/Computer_file) for information about files.",
        examples=["library"],
        title="Component Type",
    )
    mime_type: Optional[constr(regex=r"^[-+a-z0-9.]+/[-+a-z0-9.]+$")] = Field(
        None,
        alias="mime-type",
        description="The optional mime-type of the component. When used on file components, the mime-type can provide additional context about the kind of file being represented such as an image, font, or executable. Some library or framework components may also have an associated mime-type.",
        examples=["image/jpeg"],
        title="Mime-Type",
    )
    bom_ref: Optional[RefType] = Field(
        None,
        alias="bom-ref",
        description="An optional identifier which can be used to reference the component elsewhere in the BOM. Every bom-ref MUST be unique within the BOM.",
        title="BOM Reference",
    )
    supplier: Optional[OrganizationalEntity] = Field(
        None,
        description=" The organization that supplied the component. The supplier may often be the manufacturer, but may also be a distributor or repackager.",
        title="Component Supplier",
    )
    author: Optional[str] = Field(
        None,
        description="The person(s) or organization(s) that authored the component",
        examples=["Acme Inc"],
        title="Component Author",
    )
    publisher: Optional[str] = Field(
        None,
        description="The person(s) or organization(s) that published the component",
        examples=["Acme Inc"],
        title="Component Publisher",
    )
    group: Optional[str] = Field(
        None,
        description="The grouping name or identifier. This will often be a shortened, single name of the company or project that produced the component, or the source package or domain name. Whitespace and special characters should be avoided. Examples include: apache, org.apache.commons, and apache.org.",
        examples=["com.acme"],
        title="Component Group",
    )
    name: str = Field(
        ...,
        description="The name of the component. This will often be a shortened, single name of the component. Examples: commons-lang3 and jquery",
        examples=["tomcat-catalina"],
        title="Component Name",
    )
    version: Optional[str] = Field(
        None,
        description="The component version. The version should ideally comply with semantic versioning but is not enforced.",
        examples=["9.0.14"],
        title="Component Version",
    )
    description: Optional[str] = Field(
        None,
        description="Specifies a description for the component",
        title="Component Description",
    )
    scope: Optional[Scope] = Field(
        "required",
        description="Specifies the scope of the component. If scope is not specified, 'required' scope SHOULD be assumed by the consumer of the BOM.",
        title="Component Scope",
    )
    hashes: Optional[List[Hash]] = Field(None, title="Component Hashes")
    licenses: Optional[List[LicenseChoice]] = Field(None, title="Component License(s)")
    copyright: Optional[str] = Field(
        None,
        description="A copyright notice informing users of the underlying claims to copyright ownership in a published work.",
        examples=["Acme Inc"],
        title="Component Copyright",
    )
    cpe: Optional[str] = Field(
        None,
        description="Specifies a well-formed CPE name that conforms to the CPE 2.2 or 2.3 specification. See [https://nvd.nist.gov/products/cpe](https://nvd.nist.gov/products/cpe)",
        examples=["cpe:2.3:a:acme:component_framework:-:*:*:*:*:*:*:*"],
        title="Component Common Platform Enumeration (CPE)",
    )
    purl: Optional[str] = Field(
        None,
        description="Specifies the package-url (purl). The purl, if specified, MUST be valid and conform to the specification defined at: [https://github.com/package-url/purl-spec](https://github.com/package-url/purl-spec)",
        examples=["pkg:maven/com.acme/tomcat-catalina@9.0.14?packaging=jar"],
        title="Component Package URL (purl)",
    )
    swid: Optional[Swid] = Field(
        None,
        description="Specifies metadata and content for [ISO-IEC 19770-2 Software Identification (SWID) Tags](https://www.iso.org/standard/65666.html).",
        title="SWID Tag",
    )
    modified: Optional[bool] = Field(
        None,
        description="[Deprecated] - DO NOT USE. This will be removed in a future version. Use the pedigree element instead to supply information on exactly how the component was modified. A boolean value indicating if the component has been modified from the original. A value of true indicates the component is a derivative of the original. A value of false indicates the component has not been modified from the original.",
        title="Component Modified From Original",
    )
    pedigree: Optional[Pedigree] = Field(
        None,
        description="Component pedigree is a way to document complex supply chain scenarios where components are created, distributed, modified, redistributed, combined with other components, etc. Pedigree supports viewing this complex chain from the beginning, the end, or anywhere in the middle. It also provides a way to document variants where the exact relation may not be known.",
        title="Component Pedigree",
    )
    externalReferences: Optional[List[ExternalReference]] = Field(
        None,
        description="External references provide a way to document systems, sites, and information that may be relevant but which are not included with the BOM.",
        title="External References",
    )
    components: Optional[List[Component]] = Field(
        None,
        description="A list of software and hardware components included in the parent component. This is not a dependency tree. It provides a way to specify a hierarchical representation of component assemblies, similar to system &#8594; subsystem &#8594; parts assembly in physical supply chains.",
        title="Components",
    )
    evidence: Optional[ComponentEvidence] = Field(
        None,
        description="Provides the ability to document evidence collected through various forms of extraction or analysis.",
        title="Evidence",
    )
    releaseNotes: Optional[ReleaseNotes] = Field(
        None, description="Specifies optional release notes.", title="Release notes"
    )
    properties: Optional[List[Property]] = Field(
        None,
        description="Provides the ability to document properties in a name-value store. This provides flexibility to include data not officially supported in the standard without having to use additional namespaces or create extensions. Unlike key-value stores, properties support duplicate names, each potentially having different values. Property names of interest to the general public are encouraged to be registered in the [CycloneDX Property Taxonomy](https://github.com/CycloneDX/cyclonedx-property-taxonomy). Formal registration is OPTIONAL.",
        title="Properties",
    )
    signature: Optional[Signature] = Field(
        None,
        description="Enveloped signature in [JSON Signature Format (JSF)](https://cyberphone.github.io/doc/security/jsf.html).",
        title="Signature",
    )


Service.update_forward_refs()
CyclonedxSoftwareBillOfMaterialsStandard.update_forward_refs()
Metadata.update_forward_refs()
Pedigree.update_forward_refs()



"""
Base class for all collector plugins
"""
import json
import os
from abc import abstractmethod
from typing import Any, List, Optional, final

from hoppr import plugin_utils
from hoppr.base_plugins.hoppr import HopprPlugin, hoppr_process, hoppr_rerunner
from hoppr.result import Result
from hoppr.types.cred_object import CredObject
from hoppr.types.cyclonedx_1_4 import Component
from hoppr.utils import dedup_list


class CollectorPlugin(HopprPlugin):
    """
    Base class for all collector plugins
    """

    @staticmethod
    def _get_repos(comp: Component) -> List[str]:
        """
        Returns all repos listed in all "hoppr:repository:component_search_sequence" properties
        for this component
        """
        repo_list = []

        for prop in comp.properties or []:
            if prop.name == "hoppr:repository:component_search_sequence":
                search_sequence = json.loads(prop.value or "")
                repo_list.extend(search_sequence.get("Repositories", []))

        return dedup_list(repo_list)

    @abstractmethod
    @hoppr_rerunner
    def collect(self, comp: Any, repo_url: str, creds: CredObject = None):
        """
        This method should attempt to collect a single component from the specified URL
        """

    def directory_for(
        self, purl_type: str, repo_url: str, subdir: Optional[str] = None
    ) -> str:
        """
        Identify the directory into which the artifact should be copied
        """

        repo_dir = plugin_utils.dir_name_from_repo_url(repo_url)
        directory = f"{self.context.collect_root_dir}/{purl_type}/{repo_dir}/"
        if subdir is not None:
            directory += f"{subdir}/"
        os.makedirs(directory, exist_ok=True)
        return directory

    @final
    @hoppr_process
    def process_component(self, comp: Component) -> Result:
        """
        Copy a component to the local collection directory structure

        A CollectorPlugin will never return a RETRY result, but handles the retry logic internally.
        """

        logger = self.get_logger()

        result = Result.fail(f"No repository found for purl {comp.purl}")

        # for repo_url in self.context.manifest.get_repos():
        for repo_url in self._get_repos(comp):
            logger.info(  # pylint: disable=logging-fstring-interpolation
                f"Looking in repository: {repo_url}"
            )

            repo_creds = None
            if self.context.credentials is not None:
                repo_creds = self.context.credentials.find_credentials(repo_url)

            result = self.collect(comp, repo_url, repo_creds)

            if result.is_success():
                break  ### We found it, no need to try any more repositories

        return result



"""
TransferType dataclass and constructor
"""

from dataclasses import dataclass
from typing import Optional


@dataclass
class Plugin:
    """
    Plugin Dataclass containing name and optional config
    """

    name: str
    config: Optional[dict]

    def __init__(self, name: str, config: Optional[dict] = None) -> None:
        self.name = name
        self.config = config


@dataclass
class Stage:
    """
    Stage Dataclass containing an array of plugins
    """

    name: str
    plugins: list[Plugin]

    def __init__(self, name: str, plugins: list[dict]) -> None:
        self.name = name
        self.plugins = []
        for plugin in plugins:
            self.plugins.append(Plugin(**plugin))


class TransferFileContent:  # pylint: disable="too-few-public-methods"
    """
    Transfer data type class to construct TransferType Object
    """

    schema_version: str
    kind: str
    stages: list[Stage]
    max_processes: int

    def __init__(
        self,
        schemaVersion: str,  # pylint: disable="invalid-name"
        kind: str,
        stages: dict,
        max_processes: int,
    ) -> None:
        self.schema_version = schemaVersion
        self.kind = kind
        self.stages = []
        for name, stage in stages.items():
            self.stages.append(Stage(name, stage))
        self.max_processes = max_processes



"""
ManifestType dataclass and constructor
"""

from dataclasses import dataclass
from typing import Dict, List, Optional, TypeAlias

from hoppr.exceptions import HopprLoadDataError
from hoppr.types.purl_type import PurlType
from hoppr.utils import dedup_list


@dataclass
class Include:
    """
    Include dataclass containing url or local file reference
    """

    url: Optional[str]
    local: Optional[str]

    def __init__(self, url: str = None, local: str = None) -> None:
        self.url = url
        self.local = local

    def __hash__(self):
        """
        Hash must be defined to allow Include objects to be included in a set
        """
        return hash(repr(self))


@dataclass
class Metadata:
    """
    Metadata Dataclass containing name, version, description
    """

    name: str
    version: str
    description: str

    def __init__(self, name: str, version: str, description: str) -> None:
        self.name = name
        self.version = version
        self.description = description


@dataclass
class SBOMRef:
    """
    SBOM dataclass containing url or local file reference
    """

    url: Optional[str]
    local: Optional[str]

    def __init__(self, url: str = None, local: str = None) -> None:
        self.url = url
        self.local = local

    def __hash__(self):
        """
        Hash must be defined to allow SBOM objects to be included in a set
        """
        return hash(repr(self))


@dataclass
class Repository:
    """
    Repository Dataclass taking url and description
    """

    url: str
    description: str

    def __init__(self, url: str, description: str) -> None:
        self.url = url
        self.description = description

    def __hash__(self):
        """
        Hash must be defined to allow Repository objects to be included in a set
        """
        return hash(repr(self))


Repositories: TypeAlias = Dict[PurlType, List[Repository]]


class ManifestFileContent:  # pylint: disable="too-few-public-methods"
    """
    Manifest data type class to construct ManifestType Object
    """

    schema_version: str
    kind: str
    metadata: Metadata
    sbom_refs: List[SBOMRef]
    includes: List[Include]
    repositories: Repositories

    def __init__(  # pylint: disable="too-many-arguments"
        self,
        schemaVersion: str,  # pylint: disable="invalid-name"
        kind: str,
        metadata: dict,  # pylint: disable="duplicate-code"
        sboms: List[dict],
        includes: List[dict],
        repositories: dict,
    ) -> None:
        self.schema_version = schemaVersion
        self.kind = kind
        self.metadata = Metadata(**metadata)
        self.sbom_refs = dedup_list([SBOMRef(**sbom_ref) for sbom_ref in sboms])
        self.includes = dedup_list([Include(**include) for include in includes])
        self.repositories = {}
        for purl_type_enum in PurlType:
            self.repositories[purl_type_enum] = []

        for purl_type in repositories:
            for repository in repositories[purl_type]:
                try:
                    self.repositories[PurlType(purl_type)].append(
                        Repository(url=repository["url"], description=purl_type)
                    )
                except ValueError as value_error:
                    raise HopprLoadDataError(
                        f"Purl type {purl_type} is not supported"
                    ) from value_error

        for purl_type_enum in PurlType:
            self.repositories[purl_type_enum] = dedup_list(
                self.repositories[purl_type_enum]
            )



"""
Credentials business logic
"""

# Enable forward definitions
from __future__ import annotations

from os import environ
from typing import Optional

from hoppr import utils
from hoppr.exceptions import HopprCredentialsError
from hoppr.types.cred_object import CredObject
from hoppr.types.credentials_file_content import (
    CredentialFileContent,
    CredentialRequiredService,
)


class Credentials:
    """
    Credentials business logic class
    """

    def __init__(self) -> None:
        self.content = None

    @staticmethod
    def load_file(file) -> Credentials:
        """
        Creates a credentials object from a file
        """
        input_dict = utils.load_file(file)
        creds = Credentials()
        creds.populate(input_dict)

        return creds

    def find_credentials(self, url=None) -> Optional[CredObject]:
        """
        Method to find credentials that match the provided URL.
        The longest matching in the authentication object should be used.
        """

        # Find the longest "url" in the auth list that is within the "url"
        matching_service = CredentialRequiredService(url="")

        if self.content:
            for service in self.content.credential_required_services:
                if service.url in url:
                    if len(service.url) > len(matching_service.url):
                        matching_service = service

        # If no match was found, return found == False
        if matching_service.url != "":
            pass_env = matching_service.pass_env
            if pass_env not in environ:
                raise HopprCredentialsError(
                    self.content,
                    url,
                    f"'{pass_env}' not found in environment variables.",
                )

            return CredObject(
                username=str(matching_service.user), password=environ[str(pass_env)]
            )

        return None

    def populate(self, input_dict):
        """
        Populates Credentials object with dictionary contents.
        """

        self.content = CredentialFileContent(**input_dict)



"""
Collector plugin for helm charts
"""
from typing import Any, Dict, Optional

from packageurl import PackageURL  # type: ignore

from hoppr import __version__
from hoppr.base_plugins.collector import CollectorPlugin
from hoppr.base_plugins.hoppr import hoppr_rerunner
from hoppr.context import Context
from hoppr.result import Result
from hoppr.types.cred_object import CredObject


class CollectHelmPlugin(CollectorPlugin):
    """
    Class to copy helm charts
    """

    supported_purl_types = ["helm"]
    required_commands = ["helm"]

    def get_version(self) -> str:  # pylint: disable=duplicate-code
        return __version__

    def __init__(self, context: Context, config: Optional[Dict] = None) -> None:
        super().__init__(context=context, config=config)
        if self.config is not None:
            if "helm_command" in self.config:
                self.required_commands = [self.config["helm_command"]]

    @hoppr_rerunner
    def collect(
        self, comp: Any, repo_url: str, creds: CredObject = None
    ):  # pylint: disable=duplicate-code
        """
        Collect helm chart
        """

        purl = PackageURL.from_string(comp.purl)

        self.get_logger().info(  # pylint: disable=logging-fstring-interpolation
            f"Copying helm chart from {repo_url}"
        )

        source_url = f"{repo_url}/{purl.name}"
        target_dir = self.directory_for(
            purl.type, repo_url, subdir=f"{purl.name}_{purl.version}"
        )

        command = [
            "helm",
            "fetch",
            "--repo",
            source_url,
            "-d",
            target_dir,
            purl.name,
            "--version",
            purl.version,
        ]

        password_list = []
        if creds is not None:
            command = command + [
                "--username",
                creds.username,
                "--password",
                creds.password,
            ]
            password_list = [creds.password]

        result = self.run_command(command, password_list)
        if result.returncode != 0:
            # pylint: disable=duplicate-code
            msg = f"Failed to download {purl.name} version {purl.version} helm chart"
            self.get_logger().error(msg)
            return Result.retry(msg)

        self.get_logger().info(
            "Complete helm chart artifact copy for %s", purl.to_string()
        )

        return Result.success()



"""
Hoppr utility functions
"""

import importlib
import inspect
import json
from json.decoder import JSONDecodeError
from pathlib import Path
from typing import Any, List

import requests
import yaml
from yaml.parser import ParserError as YamlParserError
from yaml.scanner import ScannerError as YAMLScannerError

from hoppr.context import Context
from hoppr.exceptions import HopprLoadDataError, HopprPluginError


def _class_in_module(obj, module):
    """
    Determines if the specified object is a class
    defined in the module
    """
    if not inspect.isclass(obj):
        return False

    module_source_string = inspect.getsource(module)
    class_source_string = inspect.getsource(obj)

    if class_source_string not in module_source_string:
        return False

    return True


def plugin_instance(plugin_name, context: Context, config: Any = None):
    """
    Create an instance of an object defined by a plugin name

    Assumes the specified plugin will define exactly one concrete class, which
    will be instaniated using a default constructor (i.e., one with no parameters).
    """

    try:
        plugin = importlib.import_module(plugin_name)
    except ModuleNotFoundError as mnfe:
        raise ModuleNotFoundError(
            f"Unable to locate plug-in {plugin_name}: {mnfe}"
        ) from mnfe

    plugin_obj = None

    for name, obj in inspect.getmembers(plugin):
        if _class_in_module(obj, plugin):

            if plugin_obj is not None:
                raise HopprPluginError(
                    f"Multiple candidate classes defined in {plugin_name}: "
                    + f"{plugin_obj.__name__}, {name}"
                )

            plugin_obj = obj

    if plugin_obj is None:
        raise HopprPluginError(f"No class definition found in in {plugin_name}.")

    return plugin_obj(context=context, config=config)


def load_string(contents: str):
    """
    Try loading yaml, json, and xml files into a dict.
    """
    try:
        return_value = yaml.safe_load(contents)

        ### yaml.safe_load will sometimes return a single string rather than
        ### the required structure.  Treating that as a failure

        if not isinstance(return_value, str):
            return return_value
    except (YamlParserError, YAMLScannerError):
        pass

    try:
        return json.loads(contents)
    except JSONDecodeError:
        pass

    raise HopprLoadDataError("Unable to recognize data as either json or yaml")


def load_file(input_file_path: Path):
    """
    Load file content (either json or yml) into a dict
    """

    if not input_file_path.is_file():
        raise HopprLoadDataError(f"{input_file_path} is not a file, cannot be opened.")

    with open(input_file_path, "r", encoding="utf-8") as input_file:
        content = input_file.read()
        return load_string(content)


def load_url(url: str, auth=None):
    """
    Load config content (either json or yml) from a url into a dict
    """

    response = requests.get(url, auth=auth)
    response.raise_for_status()

    try:
        data = load_string(str(response.content))
    except HopprLoadDataError as parse_error:
        message = f"Unable to parse result from {url}."
        if response.url != url:
            message += (
                f" Request was redirected to {response.url}. "
                + "An auth issue might have occurred."
            )
        raise HopprLoadDataError(message) from parse_error

    return data


def download_file(url, dest, creds=None):
    """
    Download content from a url into a file
    """

    response = requests.get(
        url, auth=creds, allow_redirects=True, stream=True, verify=True
    )

    if 200 <= response.status_code and response.status_code < 300:
        with open(dest, "wb") as out_file:
            for chunk in response.iter_content(chunk_size=8192):
                if chunk:
                    out_file.write(chunk)

    return response


def dedup_list(list_in: List[Any]) -> List[Any]:
    """
    De-duplicate a list
    """
    return list(dict.fromkeys(list_in)) if list_in is not None else []


def obscure_passwords(command_list, lst):
    """
    Returns an input string with any specified passwords hidden
    """

    command = ""
    for arg in command_list:
        if len(command) > 0:
            command += " "
        if " " in arg:
            command += f'"{arg}"'
        else:
            command += arg

    if lst is not None:
        for password in lst:
            if password is not None:
                command = command.replace(password, "<PASSWORD HIDDEN>")

    return command



"""
Utility functions for hoppr plug-ins
"""
import urllib
from shutil import which

from hoppr.result import Result


def dir_name_from_repo_url(repo: str) -> str:
    """
    Creates a directory name based on a repo url
    """

    if repo.endswith("/") and not repo.endswith("://"):
        repo = repo[:-1]

    return urllib.parse.quote_plus(repo)


def repo_url_from_dir_name(dir_name: str) -> str:
    """
    Reconstitutes a repo url from a directory name generated by dir_name_from_repo_url
    """

    return urllib.parse.unquote_plus(dir_name)


def get_missing_commands(cmd_list):
    """
    Determine which items from a list of required commands are unavailable
    """

    missing_cmds = []

    if isinstance(cmd_list, list):
        for cmd in cmd_list:
            missing_cmds.extend(get_missing_commands(cmd))
    elif not which(cmd_list):
        missing_cmds.append(cmd_list)

    return missing_cmds


def check_for_missing_commands(cmd_list):
    """
    Raise an exception if any of a list of required commands are unavailable
    """
    missing_cmds = get_missing_commands(cmd_list)

    if len(missing_cmds) == 0:
        return Result.success()

    return Result.fail(
        f"The following required commands are unavailable: {', '.join(missing_cmds)}.  "
        + "Have them installed and try again."
    )



"""
Collector plugin for maven artifacts
"""
from tempfile import NamedTemporaryFile
from typing import Any, Dict, Optional

from packageurl import PackageURL  # type: ignore

from hoppr import __version__
from hoppr.base_plugins.collector import CollectorPlugin
from hoppr.base_plugins.hoppr import hoppr_rerunner
from hoppr.context import Context
from hoppr.result import Result
from hoppr.types.cred_object import CredObject

_DEFAULT_MAVEN_EXT = "tar.gz"
_MAVEN_DEP_PLUGIN = "org.apache.maven.plugins:maven-dependency-plugin:2.5.1"


class CollectMavenPlugin(CollectorPlugin):
    """
    Collector plugin for maven artifacts
    """

    supported_purl_types = ["maven"]
    required_commands = ["mvn"]

    def get_version(self) -> str:  # pylint: disable=duplicate-code
        return __version__

    def __init__(self, context: Context, config: Optional[Dict] = None) -> None:
        super().__init__(context=context, config=config)
        if self.config is not None:
            if "maven_command" in self.config:
                self.required_commands = [self.config["maven_command"]]

    @hoppr_rerunner
    def collect(self, comp: Any, repo_url: str, creds: Optional[CredObject] = None):
        """
        Copy a component to the local collection directory structure
        """
        purl = PackageURL.from_string(comp.purl)

        extension = purl.qualifiers.get("type", _DEFAULT_MAVEN_EXT)

        password_list = []
        if creds is not None:
            password_list = [creds.password]

        target_dir = self.directory_for(purl.type, repo_url, subdir=f"{purl.namespace}")

        self.get_logger().info(  # pylint: disable=logging-fstring-interpolation
            f"Copying maven artifact {purl.name}:{purl.version} from {repo_url} to {target_dir}"
        )

        settings_content = """<settings>
            <servers>
                <server>
                    <id>repoId</id>
                    <username>${repo.login}</username>
                    <password>${repo.pwd}</password>
                </server>
            </servers>
        </settings>
        """

        with NamedTemporaryFile(mode="w+") as settings_file:
            settings_file.write(settings_content)
            settings_file.flush()

            command = [
                self.required_commands[0],
                f"{_MAVEN_DEP_PLUGIN}:get",
                f"--settings={settings_file.name}",
                f"-DremoteRepositories=repoId::::{repo_url}",
                f"-Dpackaging={extension}",
                f"-Dartifact={purl.namespace}:{purl.name}:{purl.version}",
                f"-Ddest={target_dir}/{purl.name}-{purl.version}.{extension}",
            ]

            if creds is not None:
                command.extend(
                    [f"-Drepo.login={creds.username}", f"-Drepo.pwd={creds.password}"]
                )

            result = self.run_command(command, password_list)
            if result.returncode != 0:
                msg = f"Failed to download maven artifact {purl.namespace}:{purl.name}:{purl.version} type={extension}"
                self.get_logger().error(msg)
                return Result.retry(msg)

            if extension != "pom":
                command = [
                    self.required_commands[0],
                    f"{_MAVEN_DEP_PLUGIN}:get",
                    f"--settings={settings_file.name}",
                    f"-DremoteRepositories=repoId::::{repo_url}",
                    "-Dpackaging=pom",
                    f"-Dartifact={purl.namespace}:{purl.name}:{purl.version}",
                    f"-Ddest={target_dir}/{purl.name}-{purl.version}.pom",
                ]

                if creds is not None:
                    command.extend(
                        [
                            f"-Drepo.login={creds.username}",
                            f"-Drepo.pwd={creds.password}",
                        ]
                    )

                result = self.run_command(command, password_list)
                if result.returncode != 0:
                    msg = f"Failed to download pom for maven artifact {purl.namespace}:{purl.name}:{purl.version}"
                    self.get_logger().error(msg)
                    return Result.retry(msg)

        return Result.success()



"""
Flatten sboms
"""

# Enable forward definitions
from __future__ import annotations

import json
from typing import Optional

from packageurl import PackageURL  # type: ignore

from hoppr.configs.manifest import Manifest
from hoppr.types.cyclonedx_1_3 import (
    CyclonedxSoftwareBillOfMaterialSpecification as Bom_1_3,
)
from hoppr.types.cyclonedx_1_4 import Component as Component_1_4
from hoppr.types.cyclonedx_1_4 import (
    CyclonedxSoftwareBillOfMaterialsStandard as Bom_1_4,
)
from hoppr.types.cyclonedx_1_4 import ExternalReference, Property
from hoppr.types.manifest_file_content import Repositories
from hoppr.types.purl_type import PurlType


def to_current_component(source) -> Component_1_4:
    """
    Convert a CycloneDx component to the most current version (currently 1.4)
    """
    comp_json = source.json(by_alias=True)
    return Component_1_4.parse_raw(comp_json)


def merge_sbom(  # pylint: disable="too-many-branches"
    merged: Bom_1_4,
    to_insert: Bom_1_4 | Bom_1_3,
    to_insert_ref: Optional[str] = None,
    consolidated_repositories: Optional[Repositories] = None,
):
    """
    Add sboms
    """
    new_external_ref = None
    if to_insert_ref is not None:
        new_external_ref = ExternalReference(
            url=to_insert_ref,
            type="bom",
            comment=to_insert.serialNumber,
        )

    if to_insert.components is not None:
        if merged.components is None:
            merged.components = []

        for to_insert_comp_orig in to_insert.components:
            if to_insert_comp_orig is None or to_insert_comp_orig.purl is None:
                continue

            to_insert_comp = to_current_component(to_insert_comp_orig)

            merged_comp = None
            for comp in merged.components:
                if comp.purl is not None and comp.purl == to_insert_comp.purl:
                    merged_comp = comp

            if merged_comp is None:
                merged_comp = to_insert_comp
                merged.components.append(merged_comp)
            else:
                # TODO merge the two components # pylint: disable="fixme"
                external_ref_dict = {}
                for external_ref in (merged_comp.externalReferences or []) + (
                    to_insert_comp.externalReferences or []
                ):
                    external_ref_dict.update({external_ref.url: external_ref})

                merged_comp.externalReferences = list(external_ref_dict.values())

            if new_external_ref is not None:
                if merged_comp.externalReferences is None:
                    merged_comp.externalReferences = []
                merged_comp.externalReferences.append(new_external_ref)

            if consolidated_repositories is not None:
                purltype = PackageURL.from_string(to_insert_comp.purl).type
                search_sequence = {
                    "version": "v1",
                    "Repositories": [
                        repo.url
                        for repo in consolidated_repositories[PurlType(purltype)]
                    ],
                }
                if merged_comp.properties is None:
                    merged_comp.properties = []
                merged_comp.properties.append(
                    Property(
                        name="hoppr:repository:component_search_sequence",
                        value=json.dumps(search_sequence),
                    )
                )

    return merged


def flatten_sboms(manifest: Manifest) -> Bom_1_4:
    """
    Flatten (dedupe) sboms
    """
    flat_sbom = Bom_1_4(specVersion="1.4", version=1, bomFormat="CycloneDX")
    for index, to_insert_sbom in enumerate(manifest.sboms):

        sbom_ref = manifest.manifest_file_content.sbom_refs[index]
        if sbom_ref.local is not None:
            to_insert_sbom_ref = sbom_ref.local
        elif sbom_ref.url is not None:
            to_insert_sbom_ref = sbom_ref.url
        else:  # pragma: no cover
            to_insert_sbom_ref = None

        flat_sbom = merge_sbom(
            flat_sbom,
            to_insert_sbom,
            to_insert_sbom_ref,
            manifest.consolidated_repositories,
        )
    for child_manifest in manifest.children:
        merge_sbom(flat_sbom, flatten_sboms(child_manifest))

    return flat_sbom



"""
Collector plugin for docker images
"""
import re
import urllib
from typing import Any, Dict, Optional

from packageurl import PackageURL  # type: ignore

from hoppr import __version__
from hoppr.base_plugins.collector import CollectorPlugin
from hoppr.base_plugins.hoppr import hoppr_rerunner
from hoppr.context import Context
from hoppr.result import Result
from hoppr.types.cred_object import CredObject


class CollectDockerPlugin(CollectorPlugin):
    """
    Collector plugin for docker images
    """

    supported_purl_types = ["docker"]
    required_commands = ["skopeo"]

    def get_version(self) -> str:  # pylint: disable=duplicate-code
        return __version__

    def __init__(self, context: Context, config: Optional[Dict] = None) -> None:
        super().__init__(context=context, config=config)
        if self.config is not None:
            if "skopeo_command" in self.config:
                self.required_commands = [self.config["skopeo_command"]]

    @hoppr_rerunner
    def collect(self, comp: Any, repo_url: str, creds: CredObject = None):
        """
        Copy a component to the local collection directory structure
        """
        purl = PackageURL.from_string(comp.purl)

        source_url = re.sub(
            "([^:])(//+)",
            r"\1/",
            f"{repo_url}/{purl.namespace}/{purl.name}:{purl.version}",
        )
        source_url = re.sub(r"^https?://", "", source_url)
        if not source_url.startswith("docker://"):
            source_url = "docker://" + source_url

        file_name = urllib.parse.unquote(purl.name)
        target_dir = self.directory_for(purl.type, repo_url, subdir=purl.namespace)
        destination = f"docker-archive:{target_dir}/{file_name}_{purl.version}"

        self.get_logger().info(  # pylint: disable=logging-fstring-interpolation
            f"Copying docker image from {source_url} to {destination}"
        )

        command = [self.required_commands[0], "copy"]

        password_list = []
        if creds is not None:
            password_list = [creds.password]
            command.extend(["--src-creds", f"{creds.username}:{creds.password}"])

        if re.match("^http://", repo_url):
            command.append("--src-tls-verify=false")

        command.extend([source_url, destination])

        proc = self.run_command(command, password_list)

        if proc.returncode != 0:
            msg = (
                f"Skopeo failed to copy docker image to {destination}, "
                + f"return_code={proc.returncode}"
            )
            self.get_logger().error(msg)
            return Result.retry(msg)

        return Result.success()



"""
Class to store results of processes
"""

import enum

import requests


class ResultStatus(enum.IntEnum):
    """
    Enumeration of possible result states
    """

    SUCCESS = 0
    RETRY = 1
    FAIL = 2
    SKIP = 3


class Result:
    """
    Class to store results of processes
    """

    def __init__(self, status: ResultStatus, message=""):
        self.status = status
        self.message = message

    def __str__(self):
        return f"{self.status.name} {self.message}".strip()

    @staticmethod
    def success(message=""):
        """
        Convenience method for generating success messages
        """
        return Result(ResultStatus.SUCCESS, message)

    @staticmethod
    def retry(message=""):
        """
        Convenience method for generating retry messages
        """
        return Result(ResultStatus.RETRY, message)

    @staticmethod
    def fail(message=""):
        """
        Convenience method for generating failure messages
        """
        return Result(ResultStatus.FAIL, message)

    @staticmethod
    def skip(message=""):
        """
        Convenience method for generating skip messages
        """
        return Result(ResultStatus.SKIP, message)

    def is_success(self):
        """
        Convenience method for testing for success messages
        """
        return self.status == ResultStatus.SUCCESS

    def is_retry(self):
        """
        Convenience method for testing for retry messages
        """
        return self.status == ResultStatus.RETRY

    def is_fail(self):
        """
        Convenience method for testing for failure messages
        """
        return self.status == ResultStatus.FAIL

    def is_skip(self):
        """
        Convenience method for testing for skip messages
        """
        return self.status == ResultStatus.SKIP

    def merge(self, other: "Result"):
        """
        Logically combine two Result objects
        """
        if other.is_skip():
            return

        self.status = max(self.status, other.status)
        if self.message == "":
            self.message = other.message
        else:
            self.message += "\n" + other.message

        return

    @staticmethod
    def from_http_response(response: requests.Response):
        """
        Build a Result object from an HTTP request response
        """

        match response.status_code:
            case response_code if 200 <= response_code <= 299:  # pylint: disable=used-before-assignment
                return Result.success(f"HTTP Status Code: {response.status_code}")
            case response_code if 500 <= response_code:
                status = ResultStatus.RETRY
            case _:
                status = ResultStatus.FAIL

        message = f"HTTP Status Code: {response.status_code}; {response.text}"
        return Result(status, message)



"""
Collector plugin for raw files
"""
import os
import shutil
import urllib
from pathlib import Path
from typing import Any

from packageurl import PackageURL  # type: ignore

from hoppr import __version__
from hoppr.base_plugins.collector import CollectorPlugin
from hoppr.base_plugins.hoppr import hoppr_rerunner
from hoppr.result import Result
from hoppr.types.cred_object import CredObject
from hoppr.utils import download_file


class CollectRawPlugin(CollectorPlugin):
    """
    Collector plugin for raw files
    """

    supported_purl_types = ["binary", "generic", "raw"]

    def get_version(self) -> str:
        return __version__

    @hoppr_rerunner
    def collect(self, comp: Any, repo_url: str, creds: CredObject = None):
        """
        Copy a component to the local collection directory structure
        """
        source_url = repo_url
        if not source_url.endswith("/"):
            source_url += "/"

        purl = PackageURL.from_string(comp.purl)
        subdir = None
        if purl.namespace is not None:
            source_url += urllib.parse.unquote(purl.namespace) + "/"
            subdir = urllib.parse.unquote(purl.namespace)

        target_dir = self.directory_for(purl.type, repo_url, subdir=subdir)

        file_name = urllib.parse.unquote(purl.name)

        if source_url.startswith("file://"):
            source_file = os.path.expanduser(source_url[7:] + file_name)
            if not Path(source_file).is_file():
                return Result.fail(f"Unable to locate file {source_file}")

            self.get_logger().info(  # pylint: disable=logging-fstring-interpolation
                f"Copying file from {source_file} to directory {target_dir}"
            )
            shutil.copy(source_file, target_dir)
            return Result.success()

        self.get_logger().info(  # pylint: disable=logging-fstring-interpolation
            f"Downloading from {source_url + file_name} to file {target_dir + file_name}"
        )
        download_creds = None
        if creds is not None:
            download_creds = (creds.username, creds.password)

        response = download_file(
            source_url + file_name, target_dir + file_name, download_creds
        )

        return Result.from_http_response(response)



"""
Package Exceptions
"""


class HopprError(RuntimeError):
    """
    Base exception raised within the hoppr app.
    """


class HopprPluginError(HopprError):
    """
    Exception raised for errors working with plug-ins
    """


class HopprLoadDataError(HopprError):
    """
    Exception raised for errors loading json/yml data
    """


class HopprCredentialsError(HopprError):
    """
    Exception raised for errors loading credential data
    """
